
# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Dict

from os import environ

import logging

from neuro_san.interfaces.async_agent_session import AsyncAgentSession
from neuro_san.internals.graph.registry.agent_tool_registry import AgentToolRegistry
from neuro_san.internals.interfaces.async_agent_session_factory import AsyncAgentSessionFactory
from neuro_san.internals.tool_factories.service_tool_factory_provider import ServiceToolFactoryProvider
from neuro_san.internals.interfaces.agent_tool_factory_provider import AgentToolFactoryProvider
from neuro_san.internals.interfaces.invocation_context import InvocationContext
from neuro_san.internals.run_context.utils.external_agent_parsing import ExternalAgentParsing
from neuro_san.session.async_direct_agent_session import AsyncDirectAgentSession
from neuro_san.session.async_grpc_service_agent_session import AsyncGrpcServiceAgentSession


class ExternalAgentSessionFactory(AsyncAgentSessionFactory):
    """
    Creates AgentSessions for external agents.
    """

    def __init__(self, use_direct: bool = False):
        """
        Constructor

        :param use_direct: When True, will use a Direct session for
                    external agents that would reside on the same server.
        """
        self.use_direct: bool = use_direct

    def create_session(self, agent_url: str,
                       invocation_context: InvocationContext) -> AsyncAgentSession:
        """
        :param agent_url: An url string pointing to an external agent that came from
                    a tools list in an agent spec.
        :param invocation_context: The context policy container that pertains to the invocation
                    of the agent.
        :return: An AsyncAgentSession through which communications about the external agent can be made.
        """

        agent_location: Dict[str, str] = ExternalAgentParsing.parse_external_agent(agent_url)
        session: AsyncAgentSession = self.create_session_from_location_dict(agent_location, invocation_context)
        return session

    def create_session_from_location_dict(self, agent_location: Dict[str, str],
                                          invocation_context: InvocationContext) -> AsyncAgentSession:
        """
        :param agent_location: An agent location dictionary returned by
                    ExternalAgentParsing.parse_external_agent()
        :param invocation_context: The context policy container that pertains to the invocation
                    of the agent.
        :return: An AsyncAgentSession through which communications about the external agent can be made.
        """
        if agent_location is None:
            return None

        # Create the session.
        host = agent_location.get("host")
        port = agent_location.get("port")
        agent_name = agent_location.get("agent_name")

        # Note: It's possible we might want some filtering/translation of
        #       metadata keys not unlike what we are doing for sly_data.
        metadata: Dict[str, str] = None
        if invocation_context is not None:
            metadata = invocation_context.get_metadata()

        session: AsyncAgentSession = None
        if self.use_direct and (host is None or len(host) == 0 or host == "localhost"):
            # Optimization: We want to create a different kind of session to minimize socket usage
            # and potentially relieve the direct user of the burden of having to start a server

            tool_factory: ServiceToolFactoryProvider =\
                    ServiceToolFactoryProvider.get_instance()
            tool_registry_provider: AgentToolFactoryProvider = \
                tool_factory.get_agent_tool_factory_provider(agent_name)
            tool_registry: AgentToolRegistry = tool_registry_provider.get_agent_tool_factory()
            session = AsyncDirectAgentSession(tool_registry, invocation_context, metadata=metadata)

        if session is None:
            session = AsyncGrpcServiceAgentSession(host, port, agent_name=agent_name,
                                                   metadata=metadata)

        # Quiet any logging from leaf-common grpc stuff.
        quiet_please = logging.getLogger("leaf_common.session.grpc_client_retry")
        quiet_please.setLevel(logging.WARNING)

        return session

    @staticmethod
    def get_tool_registry(agent_name: str,
                          manifest_tool_registries: Dict[str, AgentToolRegistry]) -> AgentToolRegistry:
        """
        :param agent_name: The name of the agent to use for the session.
        :param manifest_tool_registries: Dictionary of AgentToolRegistries from the manifest
        :return: The AgentToolRegistry corresponding to the agent_name
        """

        tool_registry: AgentToolRegistry = manifest_tool_registries.get(agent_name)
        if tool_registry is None:
            message = f"""
Agent named "{agent_name}" not found in manifest file: {environ.get("AGENT_MANIFEST_FILE")}.

Some things to check:
1. If the manifest file named above is None, know that the default points
   to the one provided with the neuro-san library for a smoother out-of-box
   experience.  If the agent you wanted is not part of that standard distribution,
   you need to set the AGENT_MANIFEST_FILE environment variable to point to a
   manifest.hocon file associated with your own project(s).
2. Check that the environment variable AGENT_MANIFEST_FILE is pointing to
   the manifest.hocon file that you expect and has no typos.
3. Does your manifest.hocon file contain a key for the agent specified?
4. Does the value for the key in the manifest file have a value of 'true'?
5. Does your agent name have a typo either in the hocon file or on the command line?
"""
            raise ValueError(message)

        return tool_registry




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import List

from openai.types.beta.threads.run import Run as APIRun

from neuro_san.internals.run_context.interfaces.run import Run
from neuro_san.internals.run_context.interfaces.tool_call import ToolCall
from neuro_san.internals.run_context.openai.openai_tool_call import OpenAIToolCall


class OpenAIRun(Run):
    """
    An OpenAI implementation of a Run of an assistant.
    """

    def __init__(self, openai_run: APIRun):
        """
        Constructor

        :param openai_run: The OpenAI Run on which this implementation is based
        """
        self.openai_run = openai_run

    def get_id(self) -> str:
        """
        :return: The string id of this run
        """
        return self.openai_run.id

    def requires_action(self) -> bool:
        """
        :return: True if the status of the run requires external action.
                 False otherwise
        """
        return self.openai_run.status == "requires_action"

    def is_running(self) -> bool:
        """
        :return: True if the run is in progress. False otherwise
        """
        return self.openai_run.status in ("queued", "in_progress")

    def get_tool_calls(self) -> List[ToolCall]:
        """
        :return: A list of ToolCalls.
        """
        tool_calls: List[ToolCall] = []

        openai_tool_calls = self.openai_run.required_action.submit_tool_outputs.tool_calls
        for openai_tool_call in openai_tool_calls:
            tool_call = OpenAIToolCall(openai_tool_call)
            tool_calls.append(tool_call)

        return tool_calls

    def model_dump_json(self) -> str:
        """
        :return: This object as a JSON string
        """
        return self.openai_run.model_dump_json()




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union

import json
import traceback
import uuid

from logging import Logger
from logging import getLogger

from openai import APIError

from pydantic_core import ValidationError

from langchain.agents import Agent
from langchain.agents import AgentExecutor
from langchain.agents.tool_calling_agent.base import create_tool_calling_agent
from langchain.base_language import BaseLanguageModel
from langchain.callbacks.tracers.logging import LoggingCallbackHandler
from langchain_core.callbacks.base import BaseCallbackHandler
from langchain_core.messages.ai import AIMessage
from langchain_core.messages.base import BaseMessage
from langchain_core.messages.human import HumanMessage
from langchain_core.messages.system import SystemMessage
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.runnables import Runnable
from langchain_core.tools import BaseTool

from neuro_san.internals.errors.error_detector import ErrorDetector
from neuro_san.internals.interfaces.async_agent_session_factory import AsyncAgentSessionFactory
from neuro_san.internals.interfaces.context_type_toolbox_factory import ContextTypeToolboxFactory
from neuro_san.internals.interfaces.context_type_llm_factory import ContextTypeLlmFactory
from neuro_san.internals.interfaces.invocation_context import InvocationContext
from neuro_san.internals.journals.journal import Journal
from neuro_san.internals.journals.originating_journal import OriginatingJournal
from neuro_san.internals.messages.origination import Origination
from neuro_san.internals.messages.agent_message import AgentMessage
from neuro_san.internals.messages.agent_tool_result_message import AgentToolResultMessage
from neuro_san.internals.messages.message_utils import convert_to_base_message
from neuro_san.internals.messages.message_utils import convert_to_message_tuple
from neuro_san.internals.run_context.interfaces.agent_tool_factory import AgentToolFactory
from neuro_san.internals.run_context.interfaces.run import Run
from neuro_san.internals.run_context.interfaces.run_context import RunContext
from neuro_san.internals.run_context.interfaces.tool_caller import ToolCaller
from neuro_san.internals.run_context.langchain.journaling_callback_handler import JournalingCallbackHandler
from neuro_san.internals.run_context.langchain.journaling_tools_agent_output_parser \
    import JournalingToolsAgentOutputParser
from neuro_san.internals.run_context.langchain.langchain_openai_function_tool \
    import LangChainOpenAIFunctionTool
from neuro_san.internals.run_context.langchain.langchain_run import LangChainRun
from neuro_san.internals.run_context.langchain.langchain_token_counter import LangChainTokenCounter
from neuro_san.internals.run_context.utils.external_agent_parsing import ExternalAgentParsing
from neuro_san.internals.run_context.utils.external_tool_adapter import ExternalToolAdapter


MINUTES: float = 60.0


# pylint: disable=too-many-instance-attributes
class LangChainRunContext(RunContext):
    """
    LangChain implementation on RunContext interface supporting high-level LLM usage
    This ends up being useful:
        https://python.langchain.com/docs/modules/tools/tools_as_openai_functions/
    Note that "tools" can be just a list of OpenAI functions JSON.
    """

    # pylint: disable=too-many-arguments, too-many-positional-arguments
    def __init__(self, llm_config: Dict[str, Any],
                 parent_run_context: RunContext,
                 tool_caller: ToolCaller,
                 invocation_context: InvocationContext,
                 chat_context: Dict[str, Any]):
        """
        Constructor

        :param llm_config: The default llm_config to use as an overlay
                            for the tool-specific llm_config
        :param parent_run_context: The parent RunContext that is calling this one. Can be None.
        :param tool_caller: The tool caller to use
        :param invocation_context: The context policy container that pertains to the invocation
                    of the agent.
        :param chat_context: A ChatContext dictionary that contains all the state necessary
                to carry on a previous conversation, possibly from a different server.
        """
        # This block contains top candidates for state storage that needs to be
        # retained when session_ids go away.
        self.chat_history: List[BaseMessage] = []
        self.journal: OriginatingJournal = None
        self.llm: BaseLanguageModel = None
        self.agent: Agent = None

        # This might get modified in create_resources() (for now)
        self.llm_config: Dict[str, Any] = llm_config
        self.run_id_base: str = str(uuid.uuid4())

        self.tools: List[BaseTool] = []
        self.error_detector: ErrorDetector = None
        self.recent_human_message: HumanMessage = None
        self.tool_caller: ToolCaller = tool_caller
        self.invocation_context: InvocationContext = invocation_context
        self.chat_context: Dict[str, Any] = chat_context
        self.origin: List[Dict[str, Any]] = []
        # Default logger
        self.logger: Logger = getLogger(self.__class__.__name__)

        parent_origin: List[Dict[str, Any]] = []
        if parent_run_context is not None:

            # Get other stuff from parent if not specified
            if self.invocation_context is None:
                self.invocation_context = parent_run_context.get_invocation_context()
            if self.chat_context is None:
                self.chat_context = parent_run_context.get_chat_context()
            parent_origin = parent_run_context.get_origin()

            # Initialize the origin.
            agent_name: str = tool_caller.get_name()
            origination: Origination = self.invocation_context.get_origination()
            self.origin = origination.add_spec_name_to_origin(parent_origin, agent_name)

        self.update_from_chat_context(self.chat_context)

        # Set up so local logging gives origin info.
        if self.origin is not None and len(self.origin) > 0:
            full_name: str = Origination.get_full_name_from_origin(self.origin)
            self.logger = getLogger(full_name)

        if self.invocation_context is not None:
            base_journal: Journal = self.invocation_context.get_journal()
            self.journal = OriginatingJournal(base_journal, self.origin, self.chat_history)

    async def create_resources(self, assistant_name: str,
                               instructions: str,
                               assignments: str,
                               tool_names: List[str] = None):
        """
        Creates resources for later use within the RunContext instance.
        Results are stored as a member in this instance for future use.

        Note that even though this method is labeled as async, we don't
        really do any async method calls in here for this implementation.

        :param assistant_name: String name of the assistant
        :param instructions: string instructions that are used
                    to create the assistant/agent
        :param assignments: string assignments of function parameters that are used as input
        :param tool_names: The list of registered tool names to use.
                    Default is None implying no tool is to be called.
        """
        # DEF - Remove the arg if possible
        _ = assistant_name

        # Create the list of callbacks to pass to the LLM ChatModel
        callbacks: List[BaseCallbackHandler] = [
            JournalingCallbackHandler(self.journal)
        ]
        full_name: str = Origination.get_full_name_from_origin(self.origin)

        # Consult the agent spec for level of verbosity as it pertains to callbacks.
        agent_spec: Dict[str, Any] = self.tool_caller.get_agent_tool_spec()
        verbose: Union[bool, str] = agent_spec.get("verbose", False)
        if isinstance(verbose, str) and verbose.lower() in ("extra", "logging"):
            # This particular class adds a *lot* of very detailed messages
            # to the logs.  Add this because some people are interested in it.
            callbacks.append(LoggingCallbackHandler(self.logger))

        # Now that we have a name, we can create an ErrorDetector for the output.
        self.error_detector = ErrorDetector(full_name,
                                            error_formatter_name=agent_spec.get("error_formatter"),
                                            system_error_fragments=["Agent stopped"],
                                            agent_error_fragments=agent_spec.get("error_fragments"))

        if tool_names is not None:
            for tool_name in tool_names:
                tool: Union[BaseTool | List[BaseTool]] = await self._create_base_tool(tool_name)
                if tool is not None:
                    if isinstance(tool, List):
                        self.tools.extend(tool)
                    else:
                        self.tools.append(tool)

        prompt_template: ChatPromptTemplate = await self._create_prompt_template(instructions, assignments)

        self.agent = self.create_agent_with_fallbacks(prompt_template, callbacks)

    def create_agent_with_fallbacks(self, prompt_template: ChatPromptTemplate,
                                    callbacks: List[BaseCallbackHandler]) -> Agent:
        """
        Creates an agent with potential fallback llms to use.
        :param prompt_template: The ChatPromptTemplate to use for the agent
        :param callbacks: The list of callbacks to use when creating any LLM via the factory
        :return: An Agent (Runnable)
        """
        # Initialize our return value
        agent: Agent = None

        # Get the factory we will use
        llm_factory: ContextTypeLlmFactory = self.invocation_context.get_llm_factory()

        # Prepare a list of fallbacks.  By default the llm_config itself is a single-entry fallback list.
        fallbacks: List[Dict[str, Any]] = [self.llm_config]
        fallbacks = self.llm_config.get("fallbacks", fallbacks)

        # Initialize a list of chain fallbacks. This may or may not get filled.
        chain_fallbacks: List[Runnable] = []

        # Go through the list of fallbacks in the config.
        for index, fallback in enumerate(fallbacks):

            # Create a model we might use.
            one_llm: BaseLanguageModel = llm_factory.create_llm(fallback, callbacks=callbacks)
            one_agent: Agent = self.create_agent(prompt_template, one_llm)

            if index == 0:
                # The first agent is the one we want to be our main guy.
                agent = one_agent
                # For now. Could be problems with different providers w/ token counting.
                self.llm = one_llm
            else:
                # Anything later than the first guy is considered a fallback. Add it to the list.
                chain_fallbacks.append(one_agent)

        if len(chain_fallbacks) > 0:
            # Set up fallbacks.
            # See https://python.langchain.com/docs/how_to/tools_error/#tryexcept-tool-call
            agent = agent.with_fallbacks(chain_fallbacks)

        return agent

    def create_agent(self, prompt_template: ChatPromptTemplate, llm: BaseLanguageModel) -> Agent:
        """
        Creates an agent.
        :param prompt_template: The ChatPromptTemplate to use for the agent
        :param llm: The BaseLanguageModel to use for the agent
        :return: An Agent (Runnable)
        """
        # Initialize our return value
        agent: Agent = None

        if len(self.tools) > 0:
            agent = create_tool_calling_agent(llm, self.tools, prompt_template)

            # The above call creates a chain in this order:
            #   first:  RunnablePassthrough
            #   middle: prompt
            #           llm_with_tools
            #   last:   ToolsAgentOutputParser
            #
            # ... we need to mess with that a bit

            # Replace the output parser from the call above.
            # Per empirical experience, this is "last".
            agent.last = JournalingToolsAgentOutputParser(self.journal)
        else:
            # This uses LangChain Expression Language (LCEL), which enables a functional, pipeline-style composition
            # using "|". Here, we pass `agent_scratchpad` in the input message, but since we don't explicitly assign it
            # to `intermediate_steps` (as done in `create_tool_calling_agent`), it remains unused by the prompt.
            #
            # In contrast, `create_tool_calling_agent` can be written in LCEL as
            # RunnablePassthrough | prompt | llm_with_tools | ToolsAgentOutputParser
            # where RunnablePassthrough `agent scratchpad` convert (AgentAction, tool output) tuples into ToolMessages.
            #
            # By skipping this step, our agent functions as a pure LLM-driven system with a defined role,
            # without tool invocation logic influencing its decision-making.
            agent = prompt_template | llm | JournalingToolsAgentOutputParser(self.journal)

        return agent

    async def _create_base_tool(self, name: str) -> BaseTool:
        """
        :param name: The name of the tool to create
        :return: The BaseTool associated with the name
        """

        factory: AgentToolFactory = self.tool_caller.get_factory()
        function_json: Dict[str, Any] = None

        # Check our own local factory. Most tools live in the neighborhood.
        agent_spec: Dict[str, Any] = factory.get_agent_tool_spec(name)
        if agent_spec is None:

            # See if the agent name given could reference an external agent.
            if not ExternalAgentParsing.is_external_agent(name):
                return None

            # Use the ExternalToolAdapter to get the function specification
            # from the service call to the external agent.
            # We should be able to use the same BaseTool for langchain integration
            # purposes as we do for any other tool, though.
            # Optimization:
            #   It's possible we might want to cache these results somehow to minimize
            #   network calls.
            session_factory: AsyncAgentSessionFactory = self.invocation_context.get_async_session_factory()
            adapter = ExternalToolAdapter(session_factory, name)
            try:
                function_json = await adapter.get_function_json(self.invocation_context)
            except ValueError:
                # Could not reach the server for the external agent, so tell about it
                message: str = f"Agent/tool {name} was unreachable. Not including it as a tool."
                agent_message = AgentMessage(content=message)
                await self.journal.write_message(agent_message)
                self.logger.info(message)
        else:
            toolbox: str = agent_spec.get("toolbox")
            if toolbox:
                toolbox_factory: ContextTypeToolboxFactory = self.invocation_context.get_toolbox_factory()
                try:
                    tool_from_toolbox = toolbox_factory.create_tool_from_toolbox(toolbox, agent_spec.get("args"))
                    # If the tool from toolbox is base tool or list of base tool, return the tool as is
                    # since tool's definition and args schema are predefined in these the class of the tool.
                    if isinstance(tool_from_toolbox, BaseTool) or (
                        isinstance(tool_from_toolbox, list) and
                        all(isinstance(tool, BaseTool) for tool in tool_from_toolbox)
                    ):
                        return tool_from_toolbox
                    # Otherwise, it is a shared coded tool.
                    function_json = tool_from_toolbox

                except ValueError as tool_creation_exception:
                    # There are errors in tool creation process
                    message: str = f"Failed to create Agent/tool '{name}': {tool_creation_exception}"
                    agent_message = AgentMessage(content=message)
                    await self.journal.write_message(agent_message)
                    self.logger.info(message)
                    return None
            else:
                function_json = agent_spec.get("function")

        if function_json is None:
            return None

        # In the case of an internal agent, the name passed in for lookup should be the
        # same as what is in the spec.
        if agent_spec is not None and name != agent_spec.get("name"):
            raise ValueError(f"Tool name mismatch.  name={name}  agent_spec.name={agent_spec.get("name")}")

        # In the case of external agents, if they report a name at all, they will
        # report something different that does not identify them as external.
        # Also, most internal agents do not have a name identifier on their functional
        # JSON, which is required.  Use the agent name we are using for look-up for that
        # regardless of intent.
        function_json["name"] = name

        function_tool: BaseTool = LangChainOpenAIFunctionTool.from_function_json(function_json,
                                                                                 self.tool_caller)
        return function_tool

    async def _create_prompt_template(self, instructions: str, assignments: str) -> ChatPromptTemplate:
        """
        Creates a ChatPromptTemplate given the generic instructions
        """
        # Assemble the prompt message list
        message_list: List[Tuple[str, str]] = []

        if len(self.chat_history) == 0:
            # We have not had any chat history just yet, so build it from scratch
            # Add to our own chat history which is updated in write_message()
            system_message = SystemMessage(instructions)
            await self.journal.write_message(system_message)
            message_list.append(("system", instructions))

            # If we have assignments, add them
            if assignments is not None and len(assignments) > 0:
                system_message = SystemMessage(assignments)
                await self.journal.write_message(system_message)
                message_list.append(("system", assignments))
        else:
            # Build the prompt template from previous chat history
            # Note that since all of these are from the chat history,
            # we can consider them already reported so they do not
            # need to get written to the journal.
            for index, base_message in enumerate(self.chat_history):

                message_tuple: Tuple[str, Any] = None

                if index == 0:
                    # Always use the most current instructions
                    message_tuple = ("system", instructions)
                else:
                    message_tuple = convert_to_message_tuple(base_message)

                if message_tuple is not None:
                    message_list.append(message_tuple)

        # Fill out the rest of the prompt per the docs for create_tooling_agent()
        # Note we are not write_message()-ing the chat history because that is redundant
        # Unclear if we should somehow/someplace write_message() the agent_scratchpad at all.
        message_list.extend([
            ("placeholder", "{chat_history}"),
            ("human", "{input}"),
            ("placeholder", "{agent_scratchpad}"),
        ])

        prompt: ChatPromptTemplate = ChatPromptTemplate.from_messages(message_list)

        return prompt

    async def submit_message(self, user_message: str) -> Run:
        """
        Submits a message to the model used by this instance.

        Note that even though this method is labeled as async, we don't
        really do any async method calls in here for this implementation.

        :param user_message: The message to submit
        :return: The run which is processing the assistant's message
        """
        # Contruct a human message out of the text of the user message
        # Don't add this to the chat history yet.
        try:
            self.recent_human_message = HumanMessage(user_message)
        except ValidationError as exception:
            full_name: str = Origination.get_full_name_from_origin(self.origin)
            message = f"ValidationError in {full_name} with message: {user_message}"
            raise ValueError(message) from exception

        # Create a run to return
        run = LangChainRun(self.run_id_base, self.chat_history)
        return run

    # pylint: disable=too-many-locals
    async def wait_on_run(self, run: Run, journal: Journal = None) -> Run:
        """
        Loops on the given run's status for model invokation.

        This truly is an asynchronous method.

        :param run: The run to wait on
        :param journal: The Journal which captures the "thinking" messages.
        :return: An potentially updated run
        """

        # Create an agent executor and invoke it with the most recent human message
        # as input.
        agent_spec: Dict[str, Any] = self.tool_caller.get_agent_tool_spec()

        verbose: Union[bool, str] = agent_spec.get("verbose", False)
        if isinstance(verbose, str):
            verbose = bool(verbose.lower() in ("true", "extra", "logging"))

        max_execution_seconds: float = agent_spec.get("max_execution_seconds",
                                                      2.0 * MINUTES)
        max_iterations: int = agent_spec.get("max_iterations", 20)
        agent_executor = AgentExecutor(agent=self.agent,
                                       tools=self.tools,
                                       max_execution_time=max_execution_seconds,
                                       max_iterations=max_iterations,
                                       verbose=verbose)

        run: Run = LangChainRun(self.run_id_base, self.chat_history)

        # Chat history is updated in write_message
        await self.journal.write_message(self.recent_human_message)

        inputs = {
            "chat_history": self.chat_history,
            "input": self.recent_human_message
        }
        invoke_config = {
            "configurable": {
                "session_id": run.get_id()
            }
        }

        # Attempt to count tokens/costs while invoking the agent.
        token_counter = LangChainTokenCounter(self.llm, self.invocation_context, self.journal)
        await token_counter.count_tokens(self.ainvoke(agent_executor, inputs, invoke_config))

        return run

    async def ainvoke(self, agent_executor: AgentExecutor, inputs: Dict[str, Any], invoke_config: Dict[str, Any]):
        """
        Set the agent in motion

        :param agent_executor: The AgentExecutor to invoke
        :param inputs: The inputs to the agent_executor
        :param invoke_config: The invoke_config to send to the agent_executor
        """
        return_dict: Dict[str, Any] = None
        retries: int = 3
        exception: Exception = None
        backtrace: str = None
        while return_dict is None and retries > 0:
            try:
                return_dict: Dict[str, Any] = await agent_executor.ainvoke(inputs, invoke_config)
            except APIError as api_error:
                self.logger.warning("retrying from openai.APIError")
                retries = retries - 1
                exception = api_error
                backtrace = traceback.format_exc()
            except KeyError as key_error:
                self.logger.warning("retrying from KeyError")
                retries = retries - 1
                exception = key_error
                backtrace = traceback.format_exc()
            except ValueError as value_error:
                response = str(value_error)
                find_string = "An output parsing error occurred. " + \
                              "In order to pass this error back to the agent and have it try again, " + \
                              "pass `handle_parsing_errors=True` to the AgentExecutor. " + \
                              "This is the error: Could not parse LLM output: `"
                if response.startswith(find_string):
                    # Agent is returning good stuff, but langchain is erroring out over it.
                    # From: https://github.com/langchain-ai/langchain/issues/1358#issuecomment-1486132587
                    # Per thread consensus, this is hacky and there are better ways to go,
                    # but removes immediate impediments.
                    return_dict = {
                        "output": response.removeprefix(find_string).removesuffix("`")
                    }
                else:
                    self.logger.warning("retrying from ValueError")
                    retries = retries - 1
                    exception = value_error
                    backtrace = traceback.format_exc()

        output: str = None
        if return_dict is None and exception is not None:
            output = f"Agent stopped due to exception {exception}"
        else:
            # Other keys generally available at this point from return_dict are
            # "chat_history" and "input".
            output = return_dict.get("output", "")
            backtrace = None

        output = self.error_detector.handle_error(output, backtrace)

        return_message: BaseMessage = AIMessage(output)

        # Chat history is updated in write_message
        await self.journal.write_message(return_message)

    async def get_response(self) -> List[Any]:
        """
        :return: The list of messages from the instance's thread.
        """
        # Not sure if this is the right thing, as this will be langchain-y stuff.
        return self.chat_history

    async def submit_tool_outputs(self, run: Run, tool_outputs: List[Any]) -> Run:
        """
        :param run: The run handling the execution of the assistant
        :param tool_outputs: The tool outputs to submit
        :return: A potentially updated run handle
        """
        if tool_outputs is not None and len(tool_outputs) > 0:
            for tool_output in tool_outputs:
                tool_messages: List[BaseMessage] = self.parse_tool_output(tool_output)
                if tool_messages is not None:
                    for tool_message in tool_messages:
                        # Chat history is updated in write_message()
                        await self.journal.write_message(tool_message)

        # Create a run to return
        run = LangChainRun(self.run_id_base, self.chat_history)

        return run

    def parse_tool_output(self, tool_output: Dict[str, Any]) -> List[BaseMessage]:
        """
        Parse a single tool_output dictionary for its results
        :return: A list of messages representing the output from the tool.
        """

        # Get a Message for each output and add to the chat history.
        # Assuming dictionary
        tool_chat_list_string = tool_output.get("output", None)
        if tool_chat_list_string is None:
            # Dunno what to do with None tool output
            return None
        if isinstance(tool_chat_list_string, tuple):
            # Sometimes output comes back as a tuple.
            # The output we want is the first element of the tuple.
            tool_chat_list_string = tool_chat_list_string[0]
        if not isinstance(tool_chat_list_string, str):
            self.logger.warning("Dunno what to do with %s tool output %s",
                                str(tool_chat_list_string.__class__.__name__),
                                str(tool_chat_list_string))
            return None

        # Remove bracketing quotes from within the string
        while (tool_chat_list_string[0] == '"' and tool_chat_list_string[-1] == '"') or \
              (tool_chat_list_string[0] == "'" and tool_chat_list_string[-1] == "'"):
            tool_chat_list_string = tool_chat_list_string[1:-1]

        # Remove escaping
        tool_chat_list_string = tool_chat_list_string.replace('\\"', '"')
        # Put back some escaping of double quotes in messages that are not json.
        # We have to do this because gpt-4o seems to not like json braces in its
        # input, but now we have to deal with the consequences in the output.
        # See BranchTook._get_args_value_as_string().
        tool_chat_list_string = tool_chat_list_string.replace('\\"', '\\\\\"')

        # Decode the JSON in that string now.
        tool_chat_list: List[Dict[str, Any]] = None
        try:
            tool_chat_list = json.loads(tool_chat_list_string)
        except json.decoder.JSONDecodeError as exception:
            self.logger.error("Exception: %s parsing %s", str(exception), str(tool_chat_list_string))
            raise exception

        # The tool_output seems to contain the entire chat history of
        # the call to the tool. For now just take the last one as the answer.
        tool_result_dict = tool_chat_list[-1]

        # Turn that guy into a BaseMessage
        # You might expect that this should be a ToolMessage, but making that
        # kind of conversion at this point runs into problems with OpenAI models
        # that process them.  So, to make things continue to work, report the
        # content as an AI message - as if the bot came up with the answer itself.
        tool_message = AgentToolResultMessage(content=tool_result_dict.get("content"),
                                              tool_result_origin=tool_output.get("origin"))

        # Integrate any sly data
        tool_sly_data: Dict[str, Any] = tool_output.get("sly_data")
        if tool_sly_data and tool_sly_data != self.tool_caller.sly_data:
            # We have sly data from the tool output that is not the same as our own
            # and it has data in it.  Integrate that.
            # It's possible we might need to run a SlyDataRedactor against from_download.sly_data on this.
            self.tool_caller.sly_data.update(tool_sly_data)

        return_messages: List[BaseMessage] = [tool_message]
        return return_messages

    async def delete_resources(self, parent_run_context: RunContext = None):
        """
        Cleans up the service-side resources associated with this instance
        :param parent_run_context: A parent RunContext perhaps the same instance,
                        but perhaps not.  Default is None
        """
        self.tools = []
        self.chat_history = []
        self.agent = None
        self.recent_human_message = None
        self.llm = None
        self.journal = None

    def get_agent_tool_spec(self) -> Dict[str, Any]:
        """
        :return: the dictionary describing the data-driven agent
        """
        if self.tool_caller is None:
            return None

        return self.tool_caller.get_agent_tool_spec()

    def get_invocation_context(self) -> InvocationContext:
        """
        :return: The InvocationContext policy container that pertains to the invocation
                    of the agent.
        """
        return self.invocation_context

    def get_chat_context(self) -> Dict[str, Any]:
        """
        :return: A ChatContext dictionary that contains all the state necessary
                to carry on a previous conversation, possibly from a different server.
                Can be None when a new conversation has been started.
        """
        return self.chat_context

    def get_origin(self) -> List[Dict[str, Any]]:
        """
        :return: A List of origin dictionaries indicating the origin of the run.
                The origin can be considered a path to the original call to the front-man.
                Origin dictionaries themselves each have the following keys:
                    "tool"                  The string name of the tool in the spec
                    "instantiation_index"   An integer indicating which incarnation
                                            of the tool is being dealt with.
        """
        return self.origin

    def update_invocation_context(self, invocation_context: InvocationContext):
        """
        Update internal state based on the InvocationContext instance passed in.
        :param invocation_context: The context policy container that pertains to the invocation
        """
        self.invocation_context = invocation_context

        base_journal: Journal = self.invocation_context.get_journal()
        self.journal = OriginatingJournal(base_journal, self.origin, self.chat_history)

    def update_from_chat_context(self, chat_context: Dict[str, Any]):
        """
        :param chat_context: A ChatContext dictionary that contains all the state necessary
                to carry on a previous conversation, possibly from a different server.
        """
        self.chat_context = chat_context

        if self.chat_context is None:
            return

        # See if our origin appears in the chat histories.
        # If so, get ours from there.
        empty: List[Any] = []
        chat_histories: List[Dict[str, Any]] = self.chat_context.get("chat_histories", empty)
        our_origin_str: str = Origination.get_full_name_from_origin(self.origin)
        for one_chat_history in chat_histories:

            # See if the origin matches our own
            test_origin: List[Dict[str, Any]] = one_chat_history.get("origin", empty)
            test_origin_str: str = Origination.get_full_name_from_origin(test_origin)
            if test_origin_str != our_origin_str:
                continue

            one_messages: List[Dict[str, Any]] = one_chat_history.get("messages", empty)
            if not one_messages:
                # Empty list - Nothing to convert. Use default empty list.
                break

            self.chat_history = []
            for chat_message in one_messages:
                base_message: BaseMessage = convert_to_base_message(chat_message)
                if base_message is not None:
                    self.chat_history.append(base_message)

            # Nothing left to search for
            break

    def get_journal(self) -> Journal:
        """
        :return: The Journal associated with the instance
        """
        return self.journal




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
"""
See class comment for details
"""
from typing import Any, Dict

from tornado.web import RequestHandler


class HealthCheckHandler(RequestHandler):
    """
    Handler class for API endpoint health check.
    """

    async def get(self):
        """
        Implementation of GET request handler for API health check.
        """

        try:
            result_dict: Dict[str, Any] = \
                {"service": "neuro-san agents",
                 "status": "healthy"}
            self.set_header("Content-Type", "application/json")
            self.write(result_dict)
        except Exception:  # pylint: disable=broad-exception-caught
            # Handle unexpected errors
            self.set_status(500)
            self.write({"error": "Internal server error"})
        finally:
            self.finish()

    def data_received(self, chunk):
        """
        Method overrides abstract method of RequestHandler
        with no-op implementation.
        """
        return




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import List
from typing import TypeVar

from pydantic import ConfigDict

from langchain.agents.output_parsers.tools import ToolsAgentOutputParser
from langchain_core.outputs import Generation

from neuro_san.internals.journals.journal import Journal
from neuro_san.internals.messages.agent_message import AgentMessage

# Bizarre convention from the superclass to adhere to overridden method.
T = TypeVar("T")


class JournalingToolsAgentOutputParser(ToolsAgentOutputParser):
    """
    ToolsAgentOutputParser implementation that intercepts agent-level chatter

    We use this to intercept the "Invoking <agent> with <params>" kinds of messages
    to stream them back to the client as AgentMessages.
    """

    # Declarations of member variables here satisfy Pydantic style,
    # which is a type validator that langchain is based on which
    # is able to use JSON schema definitions to validate fields.
    journal: Journal

    # This guy needs to be a pydantic class and in order to have
    # a non-pydantic Journal as a member, we need to do this.
    model_config = ConfigDict(arbitrary_types_allowed=True)

    def __init__(self, journal: Journal):
        """
        Constructor

        :param journal: The journal to write messages to
        """
        super().__init__(journal=journal)

    async def aparse_result(self, result: list[Generation], *, partial: bool = False) -> T:
        """
        Comments and method signature from superclass method.
        Async parse a list of candidate model Generations into a specific format.

        The return value is parsed from only the first Generation in the result, which
            is assumed to be the highest-likelihood Generation.

        Args:
            result: A list of Generations to be parsed. The Generations are assumed
                to be different candidate outputs for a single model input.
            partial: Whether to parse the output as a partial result. This is useful
                for parsers that can parse partial results. Default is False.

        Returns:
            Structured output.
        """
        # Do the superclass thing
        result = await super().aparse_result(result, partial=partial)

        # By empirical observation, when the result is a list, we are getting
        # the "Invoking" message, which is useful when wanting to know agent thoughts.
        # There is one AgentAction instance in the list per tool invocation,
        # and that guy's log member has the message we want to stream as
        # agent thought..
        if isinstance(result, List):
            for action in result:
                if action.log is not None and len(action.log) > 0:
                    message = AgentMessage(content=action.log.strip())
                    await self.journal.write_message(message)

        # Note: We do not care about AgentFinish
        return result




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
"""
See class comment for details
"""

from typing import Any
from typing import Dict
from typing import List

import logging

from leaf_common.config.config_filter import ConfigFilter


class NameCorrectionConfigFilter(ConfigFilter):
    """
    ConfigFilter implementation for correcting potentially invalid internal
    agent names within a single registry.
    """

    def filter_config(self, basis_config: Dict[str, Any]) \
            -> Dict[str, Any]:
        """
        Filters the given basis config.

        Ideally this would be a Pure Function in that it would not
        modify the caller's arguments so that the caller has a chance
        to decide whether to take any changes returned.

        :param basis_config: The config dictionary to act as the basis
                for filtering
        :return: A config dictionary, potentially modified as per the
                policy encapsulated by the implementation
        """

        if basis_config is None:
            return basis_config

        tools: List[Dict[str, Any]] = basis_config.get("tools")
        if tools is None or len(tools) == 0:
            # Nothing to do. Exit early.
            return basis_config

        corrections: Dict[str, str] = {}
        errors: List[str] = []

        # Loop through all the tools making corrections or logging errors.
        tools = basis_config.get("tools")
        for tool in tools:

            name = tool.get("name")
            new_name = self.validate_name(name)

            if new_name.startswith("Error: "):
                # Add to the errors
                errors.append(new_name)
            elif new_name != name:
                # Make a correction
                tool[name] = new_name
                corrections[name] = new_name

        # Make the name corrections consistent in the tool lists
        for tool in tools:
            agent_tools: List[str] = []
            agent_tools = tool.get("tools", agent_tools)
            if not isinstance(agent_tools, List) or len(agent_tools) == 0:
                # Nothing to correct
                continue

            new_agent_tools: List[str] = []
            for agent_tool in agent_tools:
                if agent_tool in corrections:
                    new_agent_tools.append(corrections.get(agent_tool))
                else:
                    new_agent_tools.append(agent_tool)

            tool["tools"] = new_agent_tools

        # Spit out information about errors
        logger = logging.getLogger(self.__class__.__name__)
        for error in errors:
            logger.error(error)

        # Spit out information about corrections
        for original, correction in corrections.items():
            logger.info("Correcting %s to %s", str(original), str(correction))

        return basis_config

    def validate_name(self, name: str) -> str:
        """
        :param name: The name of an agent/tool as given in the hocon file.
        :return: If the name itself is valid, simply return the input.
                If the given name is invalid, attempt to correct it and return
                a new string.  If it is not correctable, return a string that starts
                with "Error: " that describes the problem..
        """
        if name is None:
            # No name is not correctable
            return "Error: no name for agent/tool"

        if not isinstance(name, str):
            # Non-string names are not correctable
            return f"Error: agent/tool name must be a string {name}"

        if len(name) == 0:
            # An empty name is not correctable
            return "Error: agent/tool name cannot be empty"

        new_name: str = name
        if "/" in new_name:
            # Names may not contain '/'.
            # We reserve this character as part of a URI specification of a hierarchy
            # of agents.
            new_name = new_name.replace("/", "_")

        return new_name




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
import os

from typing import Any
from typing import Dict
from typing import List

import logging

from openai import OpenAI
from openai import AsyncOpenAI
from openai.types.beta import Assistant
from openai.types.beta import Thread
from openai.types.beta.threads import Run
from openai.types.beta.thread_create_and_run_params import ThreadMessage


class OpenAIClient:
    """
    Implementation of different interfaces for using direct OpenAI calls.
    """

    def __init__(self, client: OpenAI = None, asynchronous: bool = False):
        """
        Constructor

        :param client: The OpenAI Client
        :param asynchronous: Boolean that says whether or not to use an asynchronous
                    OpenAI client.  Default is False.
        """
        if client is None:

            # 3 clients are needed at least, otherwise a down-chain agent
            # can't make multiple calls to its down-chain
            api_key = os.environ.get("OPENAI_API_KEY")
            if api_key is None:
                raise ValueError("ERROR: Missing OPENAI_API_KEY environment variable.")

            if asynchronous:
                client = AsyncOpenAI(api_key=api_key)
            else:
                client = OpenAI(api_key=api_key)

        self.client = client
        self.logger = logging.getLogger(self.__class__.__name__)

    def is_async(self):
        """
        :return: True is the underlying client is asynchronous.
                 False otherwise
        """
        return isinstance(self.client, AsyncOpenAI)

    async def create_run(self, thread_id: str, assistant_id: str) -> Run:
        """
        Notes: Used in langchain's OpenAIAssistantRunnable.invoke()
                Runs do not have to be deleted.
        :param thread_id:
        :param assistant_id:
        :return: A run
        """
        # Makes a POST to /threads/{thread_id}/runs
        if self.is_async():
            run: Run = await self.client.beta.threads.runs.create(thread_id=thread_id,
                                                                  assistant_id=assistant_id)
        else:
            run: Run = self.client.beta.threads.runs.create(thread_id=thread_id,
                                                            assistant_id=assistant_id)
        return run

    async def submit_tool_outputs(self, thread_id: str, run_id: str, tool_outputs: List[Any]) -> Run:
        """
        Notes: Used in langchain's OpenAIAssistantRunnable.invoke()
        :param thread_id:
        :param run_id:
        :param tool_outputs: Really a run_submit_tool_outputs_params.ToolOutput
        :return: A run
        """
        # Makes a POST to /threads/{thread_id}/runs/{run_id}/submit_tool_outputs
        if self.is_async():
            run: Run = await self.client.beta.threads.runs.submit_tool_outputs(thread_id=thread_id,
                                                                               run_id=run_id,
                                                                               tool_outputs=tool_outputs)
        else:
            run: Run = self.client.beta.threads.runs.submit_tool_outputs(thread_id=thread_id,
                                                                         run_id=run_id,
                                                                         tool_outputs=tool_outputs)
        return run

    async def retrieve(self, thread_id: str, run_id: str) -> Run:
        """
        Notes: Used in langchain's OpenAIAssistantRunnable.invoke()
        :param thread_id:
        :param run_id:
        :return: A run
        """
        # Makes a GET to /threads/{thread_id}/runs/{run_id}
        if self.is_async():
            run: Run = await self.client.beta.threads.runs.retrieve(thread_id=thread_id,
                                                                    run_id=run_id)
        else:
            run: Run = self.client.beta.threads.runs.retrieve(thread_id=thread_id,
                                                              run_id=run_id)
        return run

    async def list_runs(self, thread_id: str) -> List[Any]:
        """
        :param thread_id:
        :return: A list of runs
        """
        if self.is_async():
            results = await self.client.beta.threads.runs.list(thread_id)
        else:
            results = self.client.beta.threads.runs.list(thread_id)
        return results.data

    async def create_thread(self) -> Thread:
        """
        Note: create_thread_and_run() is used in invoke()
        :return: A thread
        """
        # Makes a POST to /threads
        if self.is_async():
            thread: Thread = await self.client.beta.threads.create()
        else:
            thread: Thread = self.client.beta.threads.create()
        return thread

    async def delete_thread(self, thread_id: str):
        """
        :param thread_id: Thread id
        :return: Nothing
        """
        if self.is_async():
            await self.client.beta.threads.delete(thread_id)
        else:
            self.client.beta.threads.delete(thread_id)
        self.logger.info("Deleted thread_id %s", thread_id)

    async def create_message(self, thread_id: str, role: str, content: str) -> ThreadMessage:
        """
        Used in invoke()
        Messages do not have to be deleted
        :param thread_id: Thread id
        :param role: ???
        :param content: ???
        """
        thread_message: ThreadMessage = None

        # Makes a POST to /threads/{thread_id}/messages
        if self.is_async():
            thread_message = \
                    await self.client.beta.threads.messages.create(thread_id=thread_id,
                                                                   role=role,
                                                                   content=content)
        else:
            thread_message = \
                self.client.beta.threads.messages.create(thread_id=thread_id,
                                                         role=role,
                                                         content=content)
        return thread_message

    async def list_messages(self, thread_id: str, order: str = "asc") -> List[ThreadMessage]:
        """
        Used in invoke()
        :param thread_id: Thread id
        :param order: optional string describing the order of the messages
        :return: A list of messages (type?)
        """
        # Makes a GET to /threads/{thread_id}/messages
        if self.is_async():
            tuples = await self.client.beta.threads.messages.list(thread_id=thread_id, order=order)

            # The async API does not return a list of ThreadMessages, but instead a list
            # of Tuples.
            messages = []
            for one_tuple in tuples:
                # It has been observed empirically that the "data" tuple
                # contains a list of ThreadMessages
                if one_tuple[0] == "data":
                    message_list = one_tuple[1]
                    for message in message_list:
                        messages.append(message)
        else:
            messages = self.client.beta.threads.messages.list(thread_id=thread_id, order=order)

        return list(messages)

    async def create_assistant(self, name: str, instructions: str, model: str) -> Assistant:
        """
        used in OpenAIAssistantRunnable.create_assistant()
        :param name: string? name for the assistant
        :param instructions: ?
        :param model: ?
        :return: An assistant
        """
        # Makes a POST to /assistants
        assistant: Assistant = None
        if self.is_async():
            assistant = await self.client.beta.assistants.create(name=name,
                                                                 instructions=instructions,
                                                                 model=model)
        else:
            assistant = self.client.beta.assistants.create(name=name,
                                                           instructions=instructions,
                                                           model=model)
        return assistant

    async def update_assistant(self, assistant_id: str, tools: List[Dict[str, Any]]) -> Assistant:
        """
        Not used in langchain!
        :param assistant_id: The id of the assistant
        :param tools: A list of tools dictionaries?  (assistant_update_params.Tool)
        :return: An assistant
        """
        # Makes a POST to /assistants/{assistant_id}
        if self.is_async():
            assistant: Assistant = await self.client.beta.assistants.update(assistant_id=assistant_id,
                                                                            tools=tools)
        else:
            assistant: Assistant = self.client.beta.assistants.update(assistant_id=assistant_id,
                                                                      tools=tools)
        return assistant

    async def list_assistants(self) -> List[Assistant]:
        """
        :return: A list of assistants in use
        """
        if self.is_async():
            results = await self.client.beta.assistants.list()
        else:
            results = self.client.beta.assistants.list()
        return results.data

    async def delete_assistant(self, assistant_id: str):
        """
        :param assistant_id: The id of the assistant
        """
        if self.is_async():
            await self.client.beta.assistants.delete(assistant_id=assistant_id)
        else:
            self.client.beta.assistants.delete(assistant_id=assistant_id)
        self.logger.info("Deleted assistant_id %s", assistant_id)




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
# Needed for method referencing different instance of the same class
# See https://stackoverflow.com/questions/33533148/how-do-i-type-hint-a-method-with-the-type-of-the-enclosing-class
from __future__ import annotations

from typing import Any
from typing import Dict
from typing import List

import asyncio
import time

from leaf_common.config.dictionary_overlay import DictionaryOverlay

from neuro_san.internals.interfaces.invocation_context import InvocationContext
from neuro_san.internals.journals.journal import Journal
from neuro_san.internals.messages.message_utils import get_last_message_with_content
from neuro_san.internals.run_context.interfaces.agent_tool_factory import AgentToolFactory
from neuro_san.internals.run_context.interfaces.run import Run
from neuro_san.internals.run_context.interfaces.run_context import RunContext
from neuro_san.internals.run_context.interfaces.tool_caller import ToolCaller
from neuro_san.internals.run_context.openai.openai_client import OpenAIClient
from neuro_san.internals.run_context.openai.openai_run import OpenAIRun


# pylint: disable=too-many-instance-attributes
class OpenAIRunContext(RunContext):
    """
    RunContext implementation supporting the context/lifetime in
    which OpenAI calls are made.
    """

    # pylint: disable=too-many-arguments, too-many-positional-arguments
    def __init__(self, llm_config: Dict[str, Any], parent_run_context: RunContext,
                 tool_caller: ToolCaller, invocation_context: InvocationContext,
                 chat_context: Dict[str, Any]):
        """
        Constructor

        :param llm_config: The default llm_config to use as an overlay
                            for the tool-specific llm_config
        :param parent_run_context: The parent RunContext (if any) to pass
                             down its resources to a new RunContext created by
                             this call.
        :param tool_caller: The tool caller to use
        :param invocation_context: The InvocationContext policy container that pertains to the invocation
                    of the agent.
        :param chat_context: A ChatContext dictionary that contains all the state necessary
                to carry on a previous conversation, possibly from a different server.
                Can be None when a new conversation has been started.
        """

        # This might get modified in create_resources() (for now)
        self.llm_config: Dict[str, Any] = llm_config
        self.tool_caller: ToolCaller = tool_caller

        # Set up the connection to OpenAI
        self.openai_client: OpenAIClient = None
        if parent_run_context is None:
            self.openai_client = OpenAIClient(asynchronous=True)
        elif isinstance(parent_run_context, OpenAIRunContext):
            # Use the same client as the parent
            self.openai_client = parent_run_context.openai_client
        else:
            raise ValueError("parent_run_context must be OpenAIRunContext")

        # Other state initialized later
        self.thread_id: str = None
        self.assistant_id: str = None
        self.invocation_context: InvocationContext = invocation_context
        self.chat_context: Dict[str, Any] = chat_context
        self.journal: Journal = self.invocation_context.get_journal()

    # pylint: disable=too-many-locals
    async def create_resources(self, assistant_name: str,
                               instructions: str,
                               assignments: str,
                               tool_names: List[str] = None):
        """
        Creates the thread resource on the OpenAI service side.
        The result is stored as a member in this instance for future use.
        :param assistant_name: String name of the assistant that can show up in the
                    OpenAI web API.
        :param instructions: string instructions that are used to create the OpenAI assistant
        :param assignments: string assignments of function parameters that are used as input
        :param tool_names: The list of registered tool names to use.
                    Default is None implying no tool is to be called.
                    Note that this implementation can only handle the 1st tool in the list
        """
        # Create the config we will use from here on out
        overlayer = DictionaryOverlay()
        tool_llm_config: Dict[str, Any] = {}

        tool_name: str = None
        if tool_names is not None and len(tool_names) > 0:
            # We can only use the first tool name in this implementation for now
            tool_name = tool_names[0]

        factory: AgentToolFactory = self.tool_caller.get_factory()
        tool_spec: Dict[str, Any] = factory.get_agent_tool_spec(tool_name)

        if tool_spec is not None:
            tool_llm_config = tool_spec.get("llm_config")
        self.llm_config = overlayer.overlay(self.llm_config, tool_llm_config)

        model_name: str = self.llm_config.get("model_name")

        use_instructions: str = instructions
        if assignments is not None:
            use_instructions = use_instructions + assignments

        # Create the assistant
        assistant = await self.openai_client.create_assistant(
            name=assistant_name,
            instructions=use_instructions,
            model=model_name,
        )
        self.assistant_id = assistant.id

        # Create the thread
        thread = await self.openai_client.create_thread()
        self.thread_id = thread.id

        use_tools = [
            {"type": "code_interpreter"}
        ]
        function_json: Dict[str, Any] = None
        if tool_spec is not None:
            function_json = tool_spec.get("function")
        if function_json is not None:
            use_tools.append({
                "type": "function",
                "function": function_json
            })
        assistant = await self.openai_client.update_assistant(
            self.assistant_id,
            tools=use_tools
        )

    async def submit_message(self, user_message: str) -> Run:
        """
        Submits a message to create an OpenAI run.
        :param user_message: The message to submit
        :return: The OpenAI run which is processing the assistant's message
        """

        await self.openai_client.create_message(
            thread_id=self.thread_id, role="user", content=user_message
        )
        openai_run = await self.openai_client.create_run(
            thread_id=self.thread_id,
            assistant_id=self.assistant_id,
        )
        run = OpenAIRun(openai_run)
        return run

    async def wait_on_run(self, run: Run, journal: Journal = None) -> Run:
        """
        Loops on the given run's status for OpenAI service-side processing
        to be done.
        :param run: The OpenAI run on their servers
        :param journal: The Journal which captures the "thinking" messages.
        :return: An potentially updated Run
        """
        messages = await self.openai_client.list_messages(thread_id=self.thread_id)
        while run.is_running():
            openai_run = await self.openai_client.retrieve(
                thread_id=self.thread_id,
                run_id=run.get_id()
            )

            sleep_seconds = 0.5
            if self.openai_client.is_async():
                await asyncio.sleep(sleep_seconds)
            else:
                time.sleep(sleep_seconds)

            latest_messages = await self.openai_client.list_messages(thread_id=self.thread_id)

            last_message_with_content = get_last_message_with_content(messages)
            latest_message_with_content = get_last_message_with_content(latest_messages)
            if (latest_messages and messages and latest_message_with_content and
                (len(latest_messages) > len(messages) or
                    (len(latest_message_with_content.content[0].text.value) >
                        len(last_message_with_content.content[0].text.value)))):
                # number_of_new_messages = len(latest_messages) - len(messages)
                # new_messages = latest_messages[-number_of_new_messages:]
                # Write to the journal
                # if journal is not None:
                messages = latest_messages

            run = OpenAIRun(openai_run)

        return run

    async def get_response(self) -> List[Any]:
        """
        :return: The list of OpenAI messages from the instance's thread.
        """
        return await self.openai_client.list_messages(thread_id=self.thread_id)

    async def submit_tool_outputs(self, run: Run, tool_outputs: List[Any]) -> Run:
        """
        :param run: The OpenAI run handling the execution of the assistant
        :param tool_outputs: The tool outputs to submit
        :return: A potentially updated OpenAI Run handle
        """
        openai_run = await self.openai_client.submit_tool_outputs(
            thread_id=self.thread_id,
            run_id=run.get_id(),
            tool_outputs=tool_outputs
        )
        run = OpenAIRun(openai_run)
        return run

    async def delete_resources(self, parent_run_context: RunContext = None):
        """
        Cleans up the OpenAI service-side resources associated with this instance
        :param parent_run_context: A parent RunContext perhaps the same instance,
                        but perhaps not.  Default is None
        """
        if parent_run_context is None:
            # No parent run context. Always try to delete the resources
            if self.assistant_id is not None:
                await self.openai_client.delete_assistant(self.assistant_id)
                self.assistant_id = None
            if self.thread_id is not None:
                await self.openai_client.delete_thread(self.thread_id)
                self.thread_id = None
        else:
            # Do not delete the resources if they are the same as the parent run context
            if self.assistant_id is not None and self.assistant_id != parent_run_context.assistant_id:
                await self.openai_client.delete_assistant(self.assistant_id)
                self.assistant_id = None
            if self.thread_id is not None and self.thread_id != parent_run_context.thread_id:
                await self.openai_client.delete_thread(self.thread_id)
                self.thread_id = None

    def get_agent_tool_spec(self) -> Dict[str, Any]:
        """
        :return: the dictionary describing the data-driven agent
        """
        if self.tool_caller is None:
            return None

        return self.tool_caller.get_agent_tool_spec()

    def get_invocation_context(self) -> InvocationContext:
        """
        :return: The InvocationContext policy container that pertains to the invocation
                    of the agent.
        """
        return self.invocation_context

    def get_chat_context(self) -> Dict[str, Any]:
        """
        :return: A ChatContext dictionary that contains all the state necessary
                to carry on a previous conversation, possibly from a different server.
                Can be None when a new conversation has been started.
        """
        return self.chat_context

    def get_origin(self) -> List[Dict[str, Any]]:
        """
        :return: A List of origin dictionaries indicating the origin of the run.
                The origin can be considered a path to the original call to the front-man.
                Origin dictionaries themselves each have the following keys:
                    "tool"                  The string name of the tool in the spec
                    "instantiation_index"   An integer indicating which incarnation
                                            of the tool is being dealt with.
        """
        return []

    def update_invocation_context(self, invocation_context: InvocationContext):
        """
        Update internal state based on the InvocationContext instance passed in.
        :param invocation_context: The context policy container that pertains to the invocation
        """
        self.invocation_context = invocation_context

    def get_journal(self) -> Journal:
        """
        :return: The Journal associated with the instance
        """
        return self.journal




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import List

from neuro_san.internals.filters.chat_context_message_filter import ChatContextMessageFilter
from neuro_san.internals.filters.compound_message_filter import CompoundMessageFilter
from neuro_san.internals.filters.message_filter import MessageFilter


class MinimalMessageFilter(CompoundMessageFilter):
    """
    A CompoundMessageFilter that lets the minimal messages needed for an agent interaction
    go through.
    """

    def __init__(self):
        """
        Constructor
        """
        filters: List[MessageFilter] = [
            ChatContextMessageFilter(),
        ]
        super().__init__(filters)




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict

import copy
import json
import uuid

import grpc

from google.protobuf.json_format import MessageToDict
from google.protobuf.json_format import Parse

from leaf_server_common.server.grpc_metadata_forwarder import GrpcMetadataForwarder
from leaf_server_common.server.request_logger import RequestLogger

from neuro_san.api.grpc import concierge_pb2 as concierge_messages
from neuro_san.api.grpc import concierge_pb2_grpc

from neuro_san.service.agent_server_logging import AgentServerLogging
from neuro_san.session.direct_concierge_session import DirectConciergeSession


class ConciergeService(concierge_pb2_grpc.ConciergeServiceServicer):
    """
    A gRPC implementation of the Neuro-San Concierge Service.
    """

    def __init__(self,
                 request_logger: RequestLogger,
                 security_cfg: Dict[str, Any],
                 server_logging: AgentServerLogging):
        """
        Set the gRPC interface up for health checking so that the service
        will be opened to callers when the mesh sees it operational, if this
        is not done the mesh will treat the service instance as nonfunctional

        :param request_logger: The instance of the RequestLogger that helps
                    keep track of stats
        :param security_cfg: A dictionary of parameters used to
                        secure the TLS and the authentication of the gRPC
                        connection.  Supplying this implies use of a secure
                        GRPC Channel.  If None, uses insecure channel.
        :param server_logging: An AgentServerLogging instance initialized so that
                        spawned asynchronous threads can also properly initialize
                        their logging.
        """
        self.request_logger = request_logger
        self.security_cfg = security_cfg
        self.server_logging: AgentServerLogging = server_logging
        self.forwarder: GrpcMetadataForwarder = self.server_logging.get_forwarder()

    # pylint: disable=no-member
    def List(self, request: concierge_messages.ConciergeRequest,
             context: grpc.ServicerContext) \
            -> concierge_messages.ConciergeResponse:
        """
        Allows a client to get the description of available agents
        served by this service.

        :param request: a ConciergeRequest
        :param context: a grpc.ServicerContext
        :return: a ConciergeResponse
        """
        request_log = None
        log_marker: str = "concierge request"
        service_logging_dict: Dict[str, str] = {
            "request_id": f"server-{uuid.uuid4()}"
        }
        request_log = self.request_logger.start_request("List",
                                                        log_marker, context,
                                                        service_logging_dict)

        # Get the metadata to forward on to another service
        metadata: Dict[str, str] = copy.copy(service_logging_dict)
        metadata.update(self.forwarder.forward(context))

        # Get our args in order to pass to grpc-free session level
        request_dict: Dict[str, Any] = MessageToDict(request)

        # Delegate to Direct*Session
        session = DirectConciergeSession(metadata=metadata,
                                         security_cfg=self.security_cfg)
        response_dict = session.list(request_dict)

        # Convert the response dictionary to a grpc message
        response_string = json.dumps(response_dict)
        response = concierge_messages.ConciergeResponse()
        Parse(response_string, response)

        if request_log is not None:
            self.request_logger.finish_request("List", log_marker, request_log)

        return response




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict

from neuro_san.internals.interfaces.context_type_llm_factory import ContextTypeLlmFactory
from neuro_san.internals.run_context.langchain.default_llm_factory import DefaultLlmFactory


class MasterLlmFactory:
    """
    Creates the correct kind of ContextTypeLlmFactory
    """

    @staticmethod
    def create_llm_factory(config: Dict[str, Any] = None) -> ContextTypeLlmFactory:
        """
        Creates an appropriate ContextTypeLlmFactory

        :param config: The config dictionary which may or may not contain
                       keys for the context_type and default llm_config
        :return: A ContextTypeLlmFactory appropriate for the context_type in the config.
        """

        llm_factory: ContextTypeLlmFactory = None
        context_type: str = MasterLlmFactory.get_context_type(config)

        if context_type.startswith("openai"):
            llm_factory = None
        elif context_type.startswith("langchain"):
            llm_factory = DefaultLlmFactory()
        else:
            # Default case
            llm_factory = DefaultLlmFactory()

        return llm_factory

    @staticmethod
    def get_context_type(config: Dict[str, Any]) -> str:
        """
        :param config: The config dictionary which may or may not contain
                       keys for the context_type and default llm_config
        :return: The context type for the config
        """
        empty: Dict[str, Any] = {}
        use_config: Dict[str, Any] = config
        if use_config is None:
            use_config = empty

        # Prepare for sanity in checks below
        context_type: str = use_config.get("context_type")
        if context_type is None:
            context_type = "langchain"
        context_type = context_type.lower()

        return context_type




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Dict

from grpc import RpcMethodHandler
from grpc import unary_stream_rpc_method_handler
from grpc import unary_unary_rpc_method_handler

import neuro_san.api.grpc.agent_pb2 as agent__pb2
from neuro_san.api.grpc.agent_pb2_grpc import AgentServiceServicer


class AgentServicerToServer:
    """
    Taken from generated gRPC code from the agent_pb2_grpc.py file
    so multiple services of the same service protobuf construction can be serviced
    by the same server with a simple addition of an agent name in the gRPC path.
    """

    def __init__(self, servicer: AgentServiceServicer):
        """
        Constructor
        """
        self.servicer: AgentServiceServicer = servicer

    def build_rpc_handlers(self):
        """
        Constructs a table of RpcMethodHandlers
        to be used for an agent service
        """
        # One entry for each grpc method defined in the agent handling protobuf
        # Note that all methods (as of 8/27/2024) are unary_unary.
        # (Watch generated _grpc.py for changes).
        # pylint: disable=no-member
        rpc_method_handlers: Dict[str, RpcMethodHandler] = {
            'Function': unary_unary_rpc_method_handler(
                    self.servicer.Function,
                    request_deserializer=agent__pb2.FunctionRequest.FromString,
                    response_serializer=agent__pb2.FunctionResponse.SerializeToString,
            ),
            'Connectivity': unary_unary_rpc_method_handler(
                    self.servicer.Connectivity,
                    request_deserializer=agent__pb2.ConnectivityRequest.FromString,
                    response_serializer=agent__pb2.ConnectivityResponse.SerializeToString,
            ),
            'StreamingChat': unary_stream_rpc_method_handler(
                    self.servicer.StreamingChat,
                    request_deserializer=agent__pb2.ChatRequest.FromString,
                    response_serializer=agent__pb2.ChatResponse.SerializeToString,
            ),
        }
        return rpc_method_handlers




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict

from neuro_san.internals.graph.registry.abstract_common_defs_config_filter \
    import AbstractCommonDefsConfigFilter


class StringCommonDefsConfigFilter(AbstractCommonDefsConfigFilter):
    """
    An AbstractCommonDefsConfigFilter implementation that takes a
    agent tool registry config that may or may not contain commondefs
    definitions for strings to substitute in by key.

    For example: Say in the config there is a top-level definition:

        "commondefs": {
            "replacement_strings": {
                "foo": "bar"
            }
        }

    This ConfigFilter implementation will replace any string containing
    the string value "{foo}" with the full string substitution "bar".
    """

    def __init__(self, replacements: Dict[str, Any] = None):
        """
        Constructor

        :param replacements: An initial replacements dictionary to start out with
                whose (copied) contents will be updated with commondefs definitions
                in the basis_config during the filter_config() entry point.
                Default is None, indicating everything needed comes from the config.
        """
        super().__init__("replacement_strings", replacements)

    def make_replacements(self, source_value: Any, replacements: Dict[str, Any]) -> Any:
        """
        Make replacements per the keys and values in the replacements dictionary

        :param source_value: The value to potentially do replacements on
        :param replacements: A dictionary of string keys to their replacements
        :return: A potentially new value if some key in the replacements dictionary
                is found to trigger a replacement, otherwise, the same source_value
                that came in.
        """

        if not isinstance(source_value, str):
            # We only modify strings in this implementation.
            # Let everything else go through unadulterated.
            return source_value

        replacement_value: str = source_value

        for search, replace in replacements.items():

            if replace is None or len(replace) == 0:
                # Nothing in the dicitonary of replacements.
                # Leave the string as-is.
                # Move along. Nothing to see here.
                continue

            if not isinstance(replace, str) or not isinstance(search, str):
                # Unclear what the user is getting at. Skip.
                continue

            if search not in replacement_value:
                # Don't need to worry about this key
                continue

            # We want to replace any instance of "{<search>}" in a string
            # with the replace value.  In order to preserve the curly braces
            # in an f-string, we need to double them.
            replacement_value = replacement_value.replace(f"{{{search}}}", replace)

        return replacement_value




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
"""
See class comment for details
"""
import json
from typing import Any, Dict

from neuro_san.http_sidecar.handlers.base_request_handler import BaseRequestHandler


class OpenApiPublishHandler(BaseRequestHandler):
    """
    Handler class for neuro-san OpenAPI service spec publishing"concierge" API call.
    """

    def get(self):
        """
        Implementation of GET request handler
        for "publish my OpenAPI specification document" call.
        """
        metadata: Dict[str, Any] = self.get_metadata()
        self.logger.info(metadata, "Start GET /api/v1/docs")
        try:
            with open(self.openapi_service_spec_path, "r", encoding='utf-8') as f_out:
                result_dict: Dict[str, Any] = json.load(f_out)
            # Return json data to the HTTP client
            self.set_header("Content-Type", "application/json")
            self.write(result_dict)
        except Exception as exc:  # pylint: disable=broad-exception-caught
            self.process_exception(exc)
        finally:
            self.do_finish()
            self.logger.info(metadata, "Finish GET /api/v1/docs")




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
"""
See class comment for details
"""
import http
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple

import json
import os

import grpc

import tornado
from tornado.web import RequestHandler

from neuro_san.http_sidecar.logging.http_logger import HttpLogger
from neuro_san.http_sidecar.interfaces.agent_authorizer import AgentAuthorizer
from neuro_san.http_sidecar.interfaces.agents_updater import AgentsUpdater
from neuro_san.interfaces.async_agent_session import AsyncAgentSession
from neuro_san.interfaces.concierge_session import ConciergeSession
from neuro_san.session.async_grpc_service_agent_session import AsyncGrpcServiceAgentSession
from neuro_san.session.grpc_concierge_session import GrpcConciergeSession


class BaseRequestHandler(RequestHandler):
    """
    Abstract handler class for neuro-san API calls.
    Provides logic to inject neuro-san service specific data
    into local handler context.
    """
    grpc_to_http = {
        grpc.StatusCode.INVALID_ARGUMENT: 400,
        grpc.StatusCode.UNAUTHENTICATED: 401,
        grpc.StatusCode.PERMISSION_DENIED: 403,
        grpc.StatusCode.NOT_FOUND: 404,
        grpc.StatusCode.ALREADY_EXISTS: 409,
        grpc.StatusCode.INTERNAL: 500,
        grpc.StatusCode.UNAVAILABLE: 503,
        grpc.StatusCode.DEADLINE_EXCEEDED: 504
    }

    request_id: int = 0

    # pylint: disable=attribute-defined-outside-init
    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-positional-arguments
    def initialize(self,
                   agent_policy: AgentAuthorizer,
                   agents_updater: AgentsUpdater,
                   port: int,
                   forwarded_request_metadata: List[str],
                   openapi_service_spec_path: str):
        """
        This method is called by Tornado framework to allow
        injecting service-specific data into local handler context.
        :param agent_policy: abstract policy for agent requests
        :param agents_updater: abstract policy for updating
                               collection of agents being served
        :param port: gRPC service port.
        :param forwarded_request_metadata: request metadata to forward.
        :param openapi_service_spec_path: file path to OpenAPI service spec.
        """

        self.agent_policy = agent_policy
        self.agents_updater = agents_updater
        self.port: int = port
        self.forwarded_request_metadata: List[str] = forwarded_request_metadata
        self.openapi_service_spec_path: str = openapi_service_spec_path
        self.logger = HttpLogger(forwarded_request_metadata)
        # Set default request_id for this request handler in case we will need it:
        BaseRequestHandler.request_id += 1

        if os.environ.get("AGENT_ALLOW_CORS_HEADERS") is not None:
            self.set_header("Access-Control-Allow-Origin", "*")
            self.set_header("Access-Control-Allow-Methods", "GET, POST, OPTIONS")
            self.set_header("Access-Control-Allow-Headers", "Content-Type, Transfer-Encoding")

    def get_metadata(self) -> Dict[str, Any]:
        """
        Extract user metadata defined by self.forwarded_request_metadata list
        from incoming request.
        :return: dictionary of user request metadata; possibly empty
        """
        headers: Dict[str, Any] = self.request.headers
        result: Dict[str, Any] = {}
        for item_name in self.forwarded_request_metadata:
            if item_name in headers.keys():
                result[item_name] = headers[item_name]
            elif item_name == "request_id":
                # Generate unique id so we have some way to track this request:
                result[item_name] = f"request-{BaseRequestHandler.request_id}"
            else:
                result[item_name] = "None"
        return result

    def get_agent_grpc_session(self,
                               metadata: Dict[str, Any],
                               agent_name: str) -> AsyncAgentSession:
        """
        Build gRPC session to talk to "main" service
        :return: AgentSession to use
        """
        grpc_session: AsyncAgentSession = \
            AsyncGrpcServiceAgentSession(
                host="localhost",
                port=self.port,
                metadata=metadata,
                agent_name=agent_name)
        return grpc_session

    def get_concierge_grpc_session(self, metadata: Dict[str, Any]) -> ConciergeSession:
        """
        Build gRPC session to talk to "concierge" service
        :return: ConciergeSession to use
        """
        grpc_session: ConciergeSession = \
            GrpcConciergeSession(
                host="localhost",
                port=self.port,
                metadata=metadata)
        return grpc_session

    async def update_agents(self, metadata: Dict[str, Any]) -> bool:
        """
        Update internal agents table by executing request
        to underlying gRPC service.
        :param metadata: metadata for request from caller context.
        :return: True if update was successful
                 False otherwise
        """
        try:
            data: Dict[str, Any] = {}
            grpc_session: ConciergeSession = self.get_concierge_grpc_session(metadata)
            agents_dict: Dict[str, Any] = grpc_session.list(data)
            agents_list = agents_dict.get("agents", [])
            agents_names: List[str] = []
            for agent_dict in agents_list:
                agents_names.append(agent_dict["agent_name"])
            self.agents_updater.update_agents(agents_names)
            return True
        except Exception as exc:  # pylint: disable=broad-exception-caught
            self.process_exception(exc)
            return False

    def extract_grpc_error_info(self, exc: grpc.aio.AioRpcError) -> Tuple[int, str, str]:
        """
        Extract user-friendly information from gRPC exception
        :param exc: gRPC service exception
        :return: tuple of 3 values:
            corresponding HTTP error code;
            name of gRPC code;
            string with additional error details.
        """
        code = exc.code()
        http_code = BaseRequestHandler.grpc_to_http.get(code, 500)
        return http_code, code.name, exc.details()

    def process_exception(self, exc: Exception):
        """
        Process exception raised during request handling
        """
        if exc is None:
            return
        if isinstance(exc, json.JSONDecodeError):
            # Handle invalid JSON input
            self.set_status(400)
            self.write({"error": "Invalid JSON format"})
            self.logger.error(self.get_metadata(), "error: Invalid JSON format")
            return

        if isinstance(exc, grpc.aio.AioRpcError):
            http_status, err_name, err_details =\
                self.extract_grpc_error_info(exc)
            self.set_status(http_status)
            err_msg: str = f"status: {http_status} grpc: {err_name} details: {err_details}"
            self.write({"error": err_msg})
            self.logger.error(self.get_metadata(), "Http server error: %s", err_msg)
            return

        # General exception case:
        self.set_status(500)
        self.write({"error": "Internal server error"})
        self.logger.error(self.get_metadata(), "Internal server error: %s", str(exc))

    def data_received(self, chunk):
        """
        Method overrides abstract method of RequestHandler
        with no-op implementation.
        """
        return

    def prepare(self):
        self.logger.info(self.get_metadata(), f"[REQUEST RECEIVED] {self.request.method} {self.request.uri}")

    def do_finish(self):
        """
        Wrapper for finish() call
        with check for closed client connection.
        """
        try:
            self.finish()
        except tornado.iostream.StreamClosedError:
            self.logger.warning(self.get_metadata(), "Finish: client closed connection unexpectedly.")

    async def do_flush(self) -> bool:
        """
        Wrapper for flush() call
        with check for closed client connection.
        """
        try:
            await self.flush()
            return True
        except tornado.iostream.StreamClosedError:
            self.logger.warning(self.get_metadata(), "Flush: client closed connection unexpectedly.")
            return False

    async def options(self, *_args, **_kwargs):
        """
        Handles OPTIONS requests for CORS support
        """
        # No body needed. Tornado will return a 204 No Content by default
        self.set_status(http.HTTPStatus.NO_CONTENT)
        self.do_finish()



# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Union

from datetime import datetime

from leaf_common.parsers.dictionary_extractor import DictionaryExtractor
from leaf_common.persistence.easy.easy_hocon_persistence import EasyHoconPersistence

from neuro_san.client.agent_session_factory import AgentSessionFactory
from neuro_san.client.streaming_input_processor import StreamingInputProcessor
from neuro_san.interfaces.agent_session import AgentSession
from neuro_san.internals.utils.file_of_class import FileOfClass
from neuro_san.message_processing.basic_message_processor import BasicMessageProcessor
from neuro_san.session.direct_agent_session import DirectAgentSession
from neuro_san.test.driver.assert_capture import AssertCapture
from neuro_san.test.evaluators.agent_evaluator_factory import AgentEvaluatorFactory
from neuro_san.test.interfaces.agent_evaluator import AgentEvaluator
from neuro_san.test.interfaces.assert_forwarder import AssertForwarder


class DataDrivenAgentTestDriver:
    """
    Class which manages the execution of a single data-driven test case
    specified as a hocon file.
    """

    TEST_KEYS: List[str] = ["text", "sly_data"]

    def __init__(self, asserts: AssertForwarder, fixtures: FileOfClass = None):
        """
        Constructor
        :param asserts: The AssertForwarder instance to use to integrate failures
                        back into the test system.
        :param fixtures: Optional path to the fixtures root.
        """
        self.asserts_basis: AssertForwarder = asserts
        self.fixtures: FileOfClass = fixtures

    def one_test(self, hocon_file: str):
        """
        Use a single hocon file in the fixtures as a test case"

        :param hocon_file: The name of the hocon from the fixtures directory.
        """
        test_case: Dict[str, Any] = self.parse_hocon_test_case(hocon_file)

        agent: str = test_case.get("agent")
        self.asserts_basis.assertIsNotNone(agent)

        # Get the success ration
        success_ratio: str = test_case.get("success_ratio", "1/1")
        self.asserts_basis.assertIn("/", success_ratio)

        # Find the integer components of the success ratio
        success_split: List[str] = success_ratio.split("/")
        num_need_success: int = int(success_split[0])
        num_iterations: int = int(success_split[-1])

        # Put some bounds on the number of iterations
        num_iterations = max(1, num_iterations)
        num_need_success = min(num_need_success, num_iterations)

        # Capture asserts for each iteration
        iteration_asserts: List[AssertCapture] = []

        # Loop through each iteration, capturing any asserts.
        num_successful: int = 0
        for index in range(num_iterations):

            _ = index

            # Capture the asserts for this iteration and add it to the list for later
            assert_capture = AssertCapture(self.asserts_basis)
            iteration_asserts.append(assert_capture)

            # Perform a single iteration of the test.
            self.one_iteration(test_case, assert_capture)

            # Update our counter if this iteration is successful
            asserts: List[AssertionError] = assert_capture.get_asserts()
            if len(asserts) > 0:
                # Not successful
                continue

            num_successful += 1
            if num_successful == num_need_success:
                # Don't do more tests than we actually need to
                break

        # Don't bother reporting any asserts if we have met our success ratio.
        # Return early to pass this test.
        if num_successful >= num_need_success:
            return

        # Find the first assert that fails and use it to fail this test
        for assert_capture in iteration_asserts:
            asserts: List[AssertionError] = assert_capture.get_asserts()
            if len(asserts) > 0:
                one_assert: AssertionError = asserts[0]
                message: str = f"""
{num_successful} of {num_iterations} iterations on agent {agent} were successful.
Need at least {num_need_success} to consider {hocon_file} test to be successful.
"""
                raise AssertionError(message) from one_assert

    def one_iteration(self, test_case: Dict[str, Any], asserts: AssertForwarder):
        """
        Perform a single iteration on the test case.

        :param test_case: The dictionary describing the data-driven test case
        :param asserts: The AssertForwarder to send asserts to.
        """

        # Get the agent to use
        agent: str = test_case.get("agent")

        # Get the connection type
        connections: Union[List[str], str] = test_case.get("connections")
        if connections is None:
            # Assume direct if not specified
            connections = ["direct"]
        elif isinstance(connections, str):
            # Make single strings into a list for consistent parsing
            connections = [connections]
        asserts.assertIsInstance(connections, List)
        asserts.assertGreater(len(connections), 0)

        # Collect the interations to test for
        empty: List[Any] = []
        interactions: List[Dict[str, Any]] = test_case.get("interactions", empty)
        asserts.assertGreater(len(interactions), 0)

        # Collect other session information
        use_direct: bool = test_case.get("use_direct", False)
        metadata: Dict[str, Any] = test_case.get("metadata", None)

        for connection in connections:

            session: AgentSession = AgentSessionFactory().create_session(connection,
                                                                         agent,
                                                                         use_direct=use_direct,
                                                                         metadata=metadata)
            chat_context: Dict[str, Any] = None
            for interaction in interactions:
                if isinstance(session, DirectAgentSession):
                    session.reset()
                chat_context = self.interact(agent, session, interaction, chat_context, asserts)

    def parse_hocon_test_case(self, hocon_file: str) -> Dict[str, Any]:
        """
        Use a single hocon file in the fixtures as a test case"

        :param hocon_file: The name of the hocon from the fixtures directory.
        """
        test_path: str = hocon_file
        if self.fixtures is not None:
            test_path = self.fixtures.get_file_in_basis(hocon_file)
        hocon = EasyHoconPersistence(must_exist=True)
        test_case: Dict[str, Any] = hocon.restore(file_reference=test_path)
        return test_case

    # pylint: disable=too-many-locals,too-many-arguments,too-many-positional-arguments
    def interact(self, agent: str, session: AgentSession, interaction: Dict[str, Any],
                 chat_context: Dict[str, Any], asserts: AssertForwarder) -> Dict[str, Any]:
        """
        Interact with an agent and evaluate its output

        :param session: The AgentSession to work with
        :param interaction: The interaction dictionary to base evalaution off of.
        :param chat_context: The chat context to use with the interaction (if any)
        :param asserts: The AssertForwarder to send asserts to.
        """
        _ = agent       # For now
        empty: Dict[str, Any] = {}

        # Prepare the processor
        now = datetime.now()
        datestr: str = now.strftime("%Y-%m-%d-%H:%M:%S")
        thinking_dir: str = f"/tmp/agent_test/{datestr}_agent"
        input_processor = StreamingInputProcessor("", None, session, thinking_dir)
        processor: BasicMessageProcessor = input_processor.get_message_processor()

        # Prepare the request
        text: str = interaction.get("text")
        sly_data: str = interaction.get("sly_data")
        chat_filter: Dict[str, Any] = {
            "chat_filter_type": interaction.get("chat_filter", "MINIMAL")
        }
        request: Dict[str, Any] = input_processor.formulate_chat_request(text, sly_data, chat_context, chat_filter)

        # Call streaming_chat()
        chat_responses: Generator[Dict[str, Any], None, None] = session.streaming_chat(request)
        for chat_response in chat_responses:
            message = chat_response.get("response", empty)
            processor.process_message(message, chat_response.get("type"))

        # Evaluate response
        response: Dict[str, Any] = interaction.get("response", empty)
        response_extractor = DictionaryExtractor(response)
        self.test_response_keys(processor, response_extractor, self.TEST_KEYS, asserts)

        # See how we should continue the conversation
        return_chat_context: Dict[str, Any] = None
        if interaction.get("continue_conversation", True):
            return_chat_context = processor.get_chat_context()

        return return_chat_context

    def test_response_keys(self, processor: BasicMessageProcessor,
                           response_extractor: DictionaryExtractor,
                           keys: List[str],
                           asserts: AssertForwarder):
        """
        Tests the given response keys

        :param processor: The BasicMessageProcessor instance to query results from.
        :param response_extractor: The DictionaryExtractor for the test structure from the test hocon file.
        :param keys: The response keys to test
        :param asserts: The AssertForwarder to send asserts to.
        """
        deeper_test_keys: List[str] = []

        for test_key in keys:

            test_key_value: Dict[str, Any] = response_extractor.get(test_key)
            if test_key_value is None:
                # Got nothing for test_key. Nothing to see here. Please move along.
                continue

            if isinstance(test_key_value, Dict):
                # The value refers to a deeper dictionary test
                for deeper_key in test_key_value.keys():
                    deeper_test_keys.append(f"{test_key}.{deeper_key}")
            else:
                # The last part of the test_key refers to a specific evaluator type.
                split: List[str] = test_key.split(".")
                evaluator_type: str = split[-1]            # Last component of .-delimited key
                verify_key: str = ".".join(split[:-1])      # All but last component of .-delimited key
                evaluator: AgentEvaluator = AgentEvaluatorFactory.create_evaluator(asserts,
                                                                                   evaluator_type)
                if evaluator is not None:
                    evaluator.evaluate(processor, verify_key, test_key_value)

        # Recurse if there are further dictionary specs to dive into
        if len(deeper_test_keys) > 0:
            self.test_response_keys(processor, response_extractor, deeper_test_keys, asserts)




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from grpc import Channel
from grpc import UnaryUnaryMultiCallable
from grpc import UnaryStreamMultiCallable

import neuro_san.api.grpc.agent_pb2 as agent__pb2


# pylint: disable=too-many-instance-attributes
class AgentServiceStub:
    """
    The service comprises all the exchanges to the backend in support of agent services.
    """

    def __init__(self, agent_name: str = ""):
        """
        Constructor.
        """
        self._agent_name: str = agent_name

        # Stub methods. These all happen to be the same kind of method, but
        # note that thare are more defined on grpc.Channel if needed (see the source).
        # pylint: disable=invalid-name
        self.Function: UnaryUnaryMultiCallable = None
        self.Connectivity: UnaryUnaryMultiCallable = None
        self.StreamingChat: UnaryStreamMultiCallable = None

    def set_agent_name(self, agent_name: str):
        """
        Exclusively called by ForwardedAgentSession.

        :param agent_name: the agent name to set
        """
        self._agent_name = agent_name

    def get_agent_name(self) -> str:
        """
        Exclusively called by tests.
        :return: the agent_name
        """
        return self._agent_name

    def __call__(self, channel: Channel):
        """
        Because of how service stubs are used to being passed around
        like a class, we use __call__() to short circuit a constructor-like
        call to use an actual instance.
        """

        # Prepare the service name given the agent name
        service_name: str = self.prepare_service_name(self._agent_name)

        # Below comes from generated _grpc.py code for the Stub,
        # with the modification of the service name going into the args.
        # One member variable for each grpc method.
        # pylint: disable=no-member
        self.Function = channel.unary_unary(
                f"/{service_name}/Function",
                request_serializer=agent__pb2.FunctionRequest.SerializeToString,
                response_deserializer=agent__pb2.FunctionResponse.FromString,
                )
        self.Connectivity = channel.unary_unary(
                f"/{service_name}/Connectivity",
                request_serializer=agent__pb2.ConnectivityRequest.SerializeToString,
                response_deserializer=agent__pb2.ConnectivityResponse.FromString,
                )
        self.StreamingChat = channel.unary_stream(
                f"/{service_name}/StreamingChat",
                request_serializer=agent__pb2.ChatRequest.SerializeToString,
                response_deserializer=agent__pb2.ChatResponse.FromString,
                )

        return self

    @staticmethod
    def prepare_service_name(agent_name: str) -> str:
        """
        Prepares the full grpc service name given the name of the agent
        :param agent_name: The string agent name
        :return: A service name that specifies the agent name as part of the routing.
        """

        # Prepare the service name on a per-agent basis
        service_name: str = ""

        # The agent name adds the voodoo to handle the request routing for each
        # agent on the same server.
        if agent_name is not None and len(agent_name) > 0:
            if len(service_name) > 0:
                service_name += "."
            service_name += f"{agent_name}"

        # This string comes from the service definition within agent.proto
        service_name += ".AgentService"

        return service_name




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import Generator
import json
import requests


from leaf_common.time.timeout import Timeout

from neuro_san.interfaces.agent_session import AgentSession


class HttpServiceAgentSession(AgentSession):
    """
    Implementation of AgentSession that talks to an HTTP service.
    This is largely only used by command-line tests.
    """

    DEFAULT_AGENT_NAME: str = "esp_decision_assistant"

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def __init__(self, host: str = None,
                 port: str = None,
                 timeout_in_seconds: int = 30,
                 metadata: Dict[str, str] = None,
                 security_cfg: Dict[str, Any] = None,
                 umbrella_timeout: Timeout = None,
                 streaming_timeout_in_seconds: int = None,
                 agent_name: str = DEFAULT_AGENT_NAME):
        """
        Creates a AgentSession that connects to the
        Agent Service and delegates its implementations to the service.

        :param host: the service host to connect to
                        If None, will use a default
        :param port: the service port
                        If None, will use a default
        :param timeout_in_seconds: timeout to use when communicating
                        with the service
        :param metadata: A grpc metadata of key/value pairs to be inserted into
                         the header. Default is None. Preferred format is a
                         dictionary of string keys to string values.
        :param security_cfg: An optional dictionary of parameters used to
                        secure the TLS and the authentication of the gRPC
                        connection.  Supplying this implies use of a secure
                        GRPC Channel.  Default is None, uses insecure channel.
        :param umbrella_timeout: A Timeout object under which the length of all
                        looping and retries should be considered
        :param streaming_timeout_in_seconds: timeout to use when streaming to/from
                        the service. Default is None, indicating connection should
                        stay open until the (last) result is yielded.
        :param agent_name: The name of the agent to talk to
        """
        _ = umbrella_timeout
        _ = streaming_timeout_in_seconds

        self.security_cfg: Dict[str, Any] = security_cfg
        self.use_host: str = "localhost"
        if host is not None:
            self.use_host = host

        self.use_port: str = str(self.DEFAULT_HTTP_PORT)
        if port is not None:
            self.use_port = port

        self.agent_name: str = agent_name
        self.timeout_in_seconds = timeout_in_seconds
        self.metadata: Dict[str, str] = metadata

    def _get_request_path(self, function: str):
        scheme: str = "http"
        if self.security_cfg is not None:
            scheme = "https"

        return f"{scheme}://{self.use_host}:{self.use_port}/api/v1/{self.agent_name}/{function}"

    def _get_headers(self) -> Dict[str, Any]:
        """
        Get headers for any outgoing request
        """
        headers: Dict[str, Any] = self.metadata
        if headers is None:
            headers = {}
        return headers

    def help_message(self, path: str) -> str:
        """
        Method returning general help message for http connectivity problems.
        :param path: url path of a request
        :return: help message
        """
        message = f"""
        Some basic suggestions to help debug connectivity issues:
        1. Ensure the server is running and reachable:
           ping <server_address>
           curl -v <server_url>
        2. Check network issues:
           traceroute <server_address>  # Linux/macOS
           tracert <server_address>  # Windows
        3. Ensure you are using correct protocol (http/https) and port number;
        4. Run service health check:
           curl <server_url:server_port>
        5. Try testing with increased timeout;
        6. Did you misspell the agent and/or method name in your {path} request path?
        7. If working with a local docker container:
           7.1 Does your http port EXPOSEd in the Dockerfile match your value for AGENT_HTTP_PORT?
           7.2 Did you add a -p <server_port>:<server_port> to your docker run command line to map container port(s)
               to your local ones?
        8. Is the agent turned on in your manifest.hocon?
        9. If you are attempting to use https, know that the default server configurations do not
           provide any of the necessary certificates for this to work and any certs used will
           need to be well known.  If you're unfamiliar with this process, it's a big deal.
           Try regular http instead.
        """
        return message

    def function(self, request_dict: Dict[str, Any]) -> Dict[str, Any]:
        """
        :param request_dict: A dictionary version of the FunctionRequest
                    protobufs structure. Has the following keys:
                        <None>
        :return: A dictionary version of the FunctionResponse
                    protobufs structure. Has the following keys:
                "function" - the dictionary description of the function
        """
        path: str = self._get_request_path("function")
        try:
            response = requests.get(path, json=request_dict, headers=self._get_headers(),
                                    timeout=self.timeout_in_seconds)
            result_dict = json.loads(response.text)
            return result_dict
        except Exception as exc:  # pylint: disable=broad-exception-caught
            raise ValueError(self.help_message(path)) from exc

    def connectivity(self, request_dict: Dict[str, Any]) -> Dict[str, Any]:
        """
        :param request_dict: A dictionary version of the ConnectivityRequest
                    protobufs structure. Has the following keys:
                        <None>
        :return: A dictionary version of the ConnectivityResponse
                    protobufs structure. Has the following keys:
                "connectivity_info" - the list of connectivity descriptions for
                                    each node in the agent network the service
                                    wants the client ot know about.
        """
        path: str = self._get_request_path("connectivity")
        try:
            response = requests.get(path, json=request_dict, headers=self._get_headers(),
                                    timeout=self.timeout_in_seconds)
            result_dict = json.loads(response.text)
            return result_dict
        except Exception as exc:  # pylint: disable=broad-exception-caught
            raise ValueError(self.help_message(path)) from exc

    def streaming_chat(self, request_dict: Dict[str, Any]) -> Generator[Dict[str, Any], None, None]:
        """
        :param request_dict: A dictionary version of the ChatRequest
                    protobufs structure. Has the following keys:
            "user_message" - A ChatMessage dict representing the user input to the chat stream
            "chat_context" - A ChatContext dict representing the state of the previous conversation
                            (if any)
        :return: An iterator of dictionary versions of the ChatResponse
                    protobufs structure. Has the following keys:
            "response"      - An optional ChatMessage dictionary.  See chat.proto for details.

            Note that responses to the chat input might be numerous and will come as they
            are produced until the system decides there are no more messages to be sent.
        """
        path: str = self._get_request_path("streaming_chat")
        try:
            with requests.post(path, json=request_dict, headers=self._get_headers(),
                               stream=True,
                               timeout=self.timeout_in_seconds) as response:
                response.raise_for_status()

                for line in response.iter_lines(decode_unicode=True):
                    if line.strip():  # Skip empty lines
                        result_dict = json.loads(line)
                        yield result_dict
        except Exception as exc:  # pylint: disable=broad-exception-caught
            raise ValueError(self.help_message(path)) from exc



# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import List

from unittest import TestCase

from neuro_san.internals.utils.file_of_class import FileOfClass
from neuro_san.test.driver.data_driven_agent_test_driver import DataDrivenAgentTestDriver
from neuro_san.test.unittest.unit_test_assert_forwarder import UnitTestAssertForwarder


class DynamicHoconUnitTests:
    """
    Assists in putting together dynamic test cases from a list of hocon files
    that describe individual test cases digestible by DataDrivenAgentTestDriver.

    Usually a single instance is created as a constant within a TestCase class and
    the parameterized.expand() annotation is used to dynamically create individual
    test methods from list of hocon files provided to it. Since the hocon files themselves
    completelt define the test cases, their names are expected to confer information as to
    just what they are testing.

    Instance methods here help run the TestCase methods.
    The main entry point here is one_test_hocon().

    Static methods are used in leveraging the paramaterized.expand annotation
    for parameterizing unit tests. The main entry point for that is from_hocon_list().

    See test_hocons.py for an example for how to use this effectively.
    """

    def __init__(self, source_file: str, path_to_basis: str = None):
        """
        Constructor

        :param source_file: The location of the source file of the TestCase.
        :param path_to_basis: The path from the source_file to the common root
                        directory of all the hocon test cases.
        """
        self.fixture_basis = FileOfClass(source_file, path_to_basis=path_to_basis)

    def one_test_hocon(self, test_case: TestCase, test_name: str, test_hocon: str):
        """
        Entry point for parameterized test methods.
        Uses the DataDrivenAgentTestDriver to effectuate the test.

        :param test_case: The instance of the TestCase class that will house
                    the individual dynamically created test methods for each hocon file.
                    This is used as a basis for reporting asserts within the framework
                    so they are properly reported within test running infrastructure.
        :param test_name: The name of the test method which was generated by the
                    @parameterize.expand annotation.
        :param test_hocon: The hocon file to use as a data-driven test case.
                    This is expanded to use the source_file/path_to_basis information
                    passed into the constructor.
        """
        # We get the test name, but don't really use it.
        # We require it as an argument because the @parameterized.expand() annotation
        # requires the test method that it is attached to take the test name as its first argument.
        # It's an artifact of the system we are leveraging. :shrug:
        _ = test_name

        # Find a full path to the test hocon file
        test_hocon_file: str = self.fixture_basis.get_file_in_basis(test_hocon)

        # Set up the driver
        asserts = UnitTestAssertForwarder(test_case)
        driver = DataDrivenAgentTestDriver(asserts)

        # Run the test
        driver.one_test(test_hocon_file)

    @staticmethod
    def from_hocon_list(hocon_files: List[str]) -> List[List[str]]:
        """
        Main entry point intended to be used inside a @parameterized.expand annotation
        which will enumerate specific test methods based on the data generated by this
        method.

        :param hocon_files: A list of hocon files to use as test cases.
                    These are expected to exist under the path implied by
                    source_file and path_to_basis passed to the constructor.
                    Individual test names are also created from each entry in the list,
                    so it behooves the class using this to not have too deep a structure
                    or use absolute paths as input here so as not to have test method
                    names that are too unwieldy.
        :return: A list intended to be passed to an @parameterized.expand annotation that will
                spell out what that needs to create individual test methods.
        """
        expansions: List[List[str]] = []

        # Go through the list of hocon files passed in to create a single entry for
        # each test case that @parameterized.expand can digest.
        for hocon_file in hocon_files:
            one_expansion: List[str] = DynamicHoconUnitTests.parameterized_expand_entry_from_hocon(hocon_file)
            expansions.append(one_expansion)
        return expansions

    @staticmethod
    def parameterized_expand_entry_from_hocon(hocon_file: str) -> List[str]:
        """
        :param hocon_file: A single reference to a hocon file used as a test case.
        :return: A List of strings that corresponds to an entry that the
                @parameterized.expand annotation can digest to create a single test method
                as part of a greater list of these.
        """
        # Each entry consists of:
        #   *   The test name to use
        #   *   The reference to the hocon file for the test case.
        return [DynamicHoconUnitTests.test_name_from_hocon_file(hocon_file), hocon_file]

    @staticmethod
    def test_name_from_hocon_file(hocon_file: str) -> str:
        """
        :param hocon_file: A hocon file reference that will be used as a test case
        :return: A reasonable name for the test method given the file path to the hocon.
                This implies that the path listed for the hocon file in @parameterized.expand
                should be descriptive enough to tell what the test is doing.
        """
        test_name: str = hocon_file.removesuffix(".hocon")
        test_name = test_name.replace("/", "_")
        return test_name




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List
from typing import Literal
from typing import Union

from langchain_core.messages.ai import AIMessage


class AgentToolResultMessage(AIMessage):
    """
    BaseMessage implementation of a message that came as a result from a tool.
    We use AIMessage class as a basis so that langchain can interpret the content
    correctly.  The extra field that we add here is an origin list to indicate
    where the the tool result came from.
    """

    type: Literal["agent_tool_result"] = "agent_tool_result"

    def __init__(self, content: Union[str, List[Union[str, Dict]]],
                 tool_result_origin: List[Dict[str, Any]],
                 **kwargs: Any) -> None:
        """
        Pass in content as positional arg.

        Args:
            content: The string contents of the message.
            tool_result_origin: The origin describing where the tool result came from
            kwargs: Additional fields to pass to the
        """
        super().__init__(content=content, **kwargs)
        self.tool_result_origin: List[Dict[str, Any]] = tool_result_origin




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict

from neuro_san.internals.interfaces.context_type_toolbox_factory import ContextTypeToolboxFactory
from neuro_san.internals.run_context.factory.master_llm_factory import MasterLlmFactory
from neuro_san.internals.run_context.langchain.toolbox_factory import ToolboxFactory


class MasterToolboxFactory:
    """
    Creates the correct kind of ContextTypeToolboxFactory
    """

    @staticmethod
    def create_toolbox_factory(config: Dict[str, Any] = None) -> ContextTypeToolboxFactory:
        """
        Creates an appropriate ContextTypeToolboxFactory

        :param config: The config dictionary which may or may not contain
                       keys for the context_type and default toolbox_config
        :return: A ContextTypeToolboxFactory appropriate for the context_type in the config.
        """

        toolbox_factory: ContextTypeToolboxFactory = None
        context_type: str = MasterLlmFactory.get_context_type(config)

        if context_type.startswith("openai"):
            toolbox_factory = None
        elif context_type.startswith("langchain"):
            toolbox_factory = ToolboxFactory()
        else:
            # Default case
            toolbox_factory = ToolboxFactory()

        return toolbox_factory




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Dict

from neuro_san.interfaces.agent_session import AgentSession
from neuro_san.internals.interfaces.context_type_toolbox_factory import ContextTypeToolboxFactory
from neuro_san.internals.graph.registry.agent_tool_registry import AgentToolRegistry
from neuro_san.internals.interfaces.context_type_llm_factory import ContextTypeLlmFactory
from neuro_san.internals.run_context.factory.master_toolbox_factory import MasterToolboxFactory
from neuro_san.internals.run_context.factory.master_llm_factory import MasterLlmFactory
from neuro_san.internals.graph.persistence.agent_tool_registry_restorer import AgentToolRegistryRestorer
from neuro_san.internals.graph.persistence.registry_manifest_restorer import RegistryManifestRestorer
from neuro_san.internals.interfaces.agent_tool_factory_provider import AgentToolFactoryProvider
from neuro_san.internals.tool_factories.service_tool_factory_provider import ServiceToolFactoryProvider
from neuro_san.session.direct_agent_session import DirectAgentSession
from neuro_san.session.external_agent_session_factory import ExternalAgentSessionFactory
from neuro_san.session.session_invocation_context import SessionInvocationContext


class DirectAgentSessionFactory:
    """
    Sets up everything needed to use a DirectAgentSession more as a library.
    This includes:
        * Some reading of AgentToolRegistries
        * Setting up ServiceToolFactoryProvider with agent registries
          which were read in
        * Initializing an LlmFactory
    """

    def __init__(self):
        """
        Constructor
        """
        manifest_restorer = RegistryManifestRestorer()
        self.manifest_tool_registries: Dict[str, AgentToolRegistry] = manifest_restorer.restore()
        tool_factory: ServiceToolFactoryProvider =\
            ServiceToolFactoryProvider.get_instance()
        for agent_name, tool_registry in self.manifest_tool_registries.items():
            tool_factory.add_agent_tool_registry(agent_name, tool_registry)

    def create_session(self, agent_name: str, use_direct: bool = False,
                       metadata: Dict[str, str] = None) -> AgentSession:
        """
        :param agent_name: The name of the agent to use for the session.
                This name can be something in the manifest file (with no file suffix)
                or a specific full-reference to an agent network's hocon file.
        :param use_direct: When True, will use a Direct session for
                    external agents that would reside on the same server.
        :param metadata: A grpc metadata of key/value pairs to be inserted into
                         the header. Default is None. Preferred format is a
                         dictionary of string keys to string values.
        """

        tool_registry: AgentToolRegistry = self.get_agent_tool_registry(agent_name)

        llm_factory: ContextTypeLlmFactory = MasterLlmFactory.create_llm_factory()
        toolbox_factory: ContextTypeToolboxFactory = MasterToolboxFactory.create_toolbox_factory()
        # Load once now that we know what tool registry to use.
        # Include "agent_llm_info_file" from agent network hocon to llm factory.
        agent_llm_info_file = tool_registry.get_agent_llm_info_file()
        llm_factory.load(agent_llm_info_file)
        toolbox_factory.load()

        factory = ExternalAgentSessionFactory(use_direct=use_direct)
        invocation_context = SessionInvocationContext(factory, llm_factory, toolbox_factory, metadata)
        invocation_context.start()
        session: DirectAgentSession = DirectAgentSession(tool_registry=tool_registry,
                                                         invocation_context=invocation_context,
                                                         metadata=metadata)
        return session

    def get_agent_tool_registry(self, agent_name: str) -> AgentToolRegistry:
        """
        :param agent_name: The name of the agent whose AgentToolRegistry we want to get.
                This name can be something in the manifest file (with no file suffix)
                or a specific full-reference to an agent network's hocon file.
        :return: The AgentToolRegistry corresponding to that agent.
        """

        if agent_name is None or len(agent_name) == 0:
            return None

        tool_registry: AgentToolRegistry = None
        if agent_name.endswith(".hocon") or agent_name.endswith(".json"):
            # We got a specific file name
            restorer = AgentToolRegistryRestorer()
            tool_registry = restorer.restore(file_reference=agent_name)
        else:
            # Use the standard stuff available via the manifest file.
            tool_factory: ServiceToolFactoryProvider =\
                ServiceToolFactoryProvider.get_instance()
            tool_registry_provider: AgentToolFactoryProvider =\
                tool_factory.get_agent_tool_factory_provider(agent_name)
            tool_registry = tool_registry_provider.get_agent_tool_factory()

        return tool_registry



# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any

from neuro_san.test.evaluators.abstract_agent_evaluator import AbstractAgentEvaluator


class GreaterAgentEvaluator(AbstractAgentEvaluator):
    """
    AbstractAgentEvaluator implementation that looks for inequalities of values in output.
    """

    def test_one(self, verify_value: Any, test_value: Any):
        """
        :param verify_value: The value to verify against
        :param test_value: The value appearing in the test sample
        """
        if self.negate:
            self.asserts.assertLessEqual(verify_value, test_value)
        else:
            self.asserts.assertGreater(verify_value, test_value)




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List

from neuro_san.internals.run_context.interfaces.run_context import RunContext


class CallableTool:
    """
    Interface describing what a CallingTool can access
    when invoking LLM function calls.
    """

    async def build(self) -> List[Any]:
        """
        Main entry point to the class.
        :return: A List of messages produced during this process.
        """
        raise NotImplementedError

    def get_origin(self) -> List[Dict[str, Any]]:
        """
        :return: A List of origin dictionaries indicating the origin of the tool.
                The origin can be considered a path to the original call to the front-man.
                Origin dictionaries themselves each have the following keys:
                    "tool"                  The string name of the tool in the spec
                    "instantiation_index"   An integer indicating which incarnation
                                            of the tool is being dealt with.
        """
        raise NotImplementedError

    async def delete_resources(self, parent_run_context: RunContext):
        """
        Cleans up after any allocated resources on their server side.
        :param parent_run_context: The RunContext which contains the scope
                    of operation of this CallableNode
        """
        raise NotImplementedError




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from neuro_san.session.async_grpc_service_agent_session import AsyncGrpcServiceAgentSession

AsyncServiceAgentSession = AsyncGrpcServiceAgentSession




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import List

from langchain.schema import BaseMessage

from neuro_san.internals.run_context.interfaces.run import Run
from neuro_san.internals.run_context.interfaces.tool_call import ToolCall
from neuro_san.internals.run_context.langchain.langchain_tool_call import LangChainToolCall


class LangChainRun(Run):
    """
    A LangChain implementation of a Run.
    """

    def __init__(self, run_id_base: str, chat_history: List[BaseMessage],
                 tool_name: str = None, args: Any = None):
        """
        Constructor

        :param run_id_base: The basis string for run ids.
        :param chat_history: The chat history so far
        :param tool_name: The name of the tool to call
        :param args: The arguments used for the tool during the run
                    represented by this instance.
        """
        self.id: str = self._create_run_id(run_id_base, chat_history)
        self.chat_history: List[BaseMessage] = chat_history
        self.tool_name: str = tool_name
        self.args: Any = args

    @staticmethod
    def _create_run_id(run_id_base: str, chat_history: List[BaseMessage]) -> str:
        """
        :return: A run id string based on the uuid in run_id_base and
                 the length of the chat history so far.
        """
        chat_length = len(chat_history)

        # Zero pad the chat_length out to 4 digits.
        run_id: str = f"run_{run_id_base}_{chat_length:04}"
        return run_id

    def get_id(self) -> str:
        """
        :return: The string id of this run
        """
        return self.id

    def requires_action(self) -> bool:
        """
        :return: True if the status of the run requires external action.
                 False otherwise
        """
        # This allows the higher level check in branch tools
        # to know that we are doing something.
        return False

    def get_tool_calls(self) -> List[ToolCall]:
        """
        :return: A list of ToolCalls.
        """
        # At least return an empty list
        tool_calls: List[ToolCall] = []
        if self.args is not None:
            # We don't yet know how to get the list of tools an AgentExecutor actually
            # wants to call (as opposed to the list it *can( call, which is in the spec).
            # It probably has something to do with adding a CallbackManagerForChainRun
            # which somehow would get its on_agent_action() called when a tool is chosen in
            # AgentExecutor._perform_agent_action().  We'd probably need to set that up
            # and allow it to add tools/args to this run instance.
            #
            # For now, we just do a single tool given the run...
            # Create a ToolCall with the correct args and add that to the ToolCalls list
            tool_call: LangChainToolCall = LangChainToolCall(self.tool_name, self.args, self.id)
            tool_calls.append(tool_call)

        return tool_calls

    def model_dump_json(self) -> str:
        """
        :return: This object as a JSON string
        """
        return "{ " + f"'id': '{self.id}'" + " }"

    def get_chat_history(self) -> List[BaseMessage]:
        """
        This is used when the LangChainOpenAIFunctionTool wants to
        know what the most recent message was from the tool as
        its output.
        :return: The list of messages comprising the chat history
        """
        return self.chat_history



from typing import Any
from typing import Dict
from typing import Union

from neuro_san.interfaces.coded_tool import CodedTool


class Accountant(CodedTool):
    """
    A tool that updates a running cost each time it is called.
    """

    def invoke(self, args: Dict[str, Any], sly_data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Updates the passed running cost each time it's called.
        :param args: A dictionary with the following keys:
                    "running_cost": the running cost to update.

        :param sly_data: A dictionary containing parameters that should be kept out of the chat stream.
                         Keys expected for this implementation are:
                         None

        :return: A dictionary containing:
                 "running_cost": the updated running cost.
        """
        tool_name = self.__class__.__name__
        print(f"========== Calling {tool_name} ==========")
        # Parse the arguments
        print(f"args: {args}")
        running_cost: float = float(args.get("running_cost"))

        # Increment the running cost not using value other 1
        # This would make a little hard if the LLM wanted to guess
        updated_running_cost: float = running_cost + 3.0

        tool_response = {
            "running_cost": updated_running_cost
        }
        print("-----------------------")
        print(f"{tool_name} response: ", tool_response)
        print(f"========== Done with {tool_name} ==========")
        return tool_response

    async def async_invoke(self, args: Dict[str, Any], sly_data: Dict[str, Any]) -> Union[Dict[str, Any], str]:
        """
        Delegates to the synchronous invoke method because it's quick, non-blocking.
        """
        return self.invoke(args, sly_data)



# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List

from copy import copy

from neuro_san.internals.run_context.utils.external_agent_parsing import ExternalAgentParsing


class Origination:
    """
    An instance for common code and state that manipulates and keeps track of origin information
    on a per-AgentSession basis.

    A full origin description is a List of Dictionaries indicating the origin of a chat message.
    An origin can be considered a path to the original call to the front-man.
    Origin dictionaries themselves each have the following keys:
        "tool"                  The string name of the tool in the spec
        "instantiation_index"   An integer indicating which incarnation
                                of the tool is being dealt with.
    """

    INSTANTIATION_START: int = 1
    NUM_INSTANTIATION_INDEX_DIGITS: int = 2

    def __init__(self):
        """
        Constructor
        """
        self.tool_to_index_map: Dict[str, int] = {}

    def add_spec_name_to_origin(self, origin: List[Dict[str, Any]], agent_name: str) \
            -> List[Dict[str, Any]]:
        """
        Adds a single component origin dictionary to the given origin list.
        :param origin: A new List of origin dictionaries indicating the origin of the run.
                The origin can be considered a path to the original call to the front-man.
                Origin dictionaries themselves each have the following keys:
                    "tool"                  The string name of the tool in the spec
                    "instantiation_index"   An integer indicating which incarnation
                                            of the tool is being dealt with.
        :param agent_name: The agent name to be added to the list.
        :return: The new origin with the agent name at the end of the list
        """
        new_origin: List[Dict[str, Any]] = []

        # Add the name from the spec to the origin, if we have it.
        if origin is None or agent_name is None:
            return new_origin

        # Make a shallow copy of the list, as we are going to modify it
        # and don't want to muck with the original
        new_origin = copy(origin)

        # Find the current instantiation index for the tool
        # and increment it in the map for later use
        instantiation_index: int = self.tool_to_index_map.get(agent_name, Origination.INSTANTIATION_START)
        self.tool_to_index_map[agent_name] = instantiation_index + 1

        # Prepare the origin dictionary to append
        origin_dict: Dict[str, Any] = {
            "tool": agent_name,
            "instantiation_index": instantiation_index
        }
        new_origin.append(origin_dict)

        return new_origin

    @staticmethod
    def get_full_name_from_origin(origin: List[Dict[str, Any]]) -> str:
        """
        :param origin: A List of origin dictionaries indicating the origin of the run.
                The origin can be considered a path to the original call to the front-man.
                Origin dictionaries themselves each have the following keys:
                    "tool"                  The string name of the tool in the spec
                    "instantiation_index"   An integer indicating which incarnation
                                            of the tool is being dealt with.
        :return: A single string name given an origin path/list
        """
        if origin is None:
            return None

        # Connect all the elements of the origin by the delimiter "."
        origin_list: List[str] = []
        for origin_dict in origin:

            # Get basic fields from the dict
            instantiation_index: int = origin_dict.get("instantiation_index", Origination.INSTANTIATION_START)
            tool: str = origin_dict.get("tool")
            if tool is None:
                # No information of value will be conveyed with no tool set in the dict.
                raise ValueError("tool name in origin_dict is None")

            # Figure out how we will deal with the index
            index_str: str = ""
            if instantiation_index > Origination.INSTANTIATION_START:
                # zfill() adds leading 0's up to the number of characters provided
                index_str = f"-{str(instantiation_index).zfill(Origination.NUM_INSTANTIATION_INDEX_DIGITS)}"

            safe_tool: str = ExternalAgentParsing.get_safe_agent_name(tool)

            # Figure out the single origin string
            origin_str: str = f"{safe_tool}{index_str}"
            origin_list.append(origin_str)

        full_path: str = ".".join(origin_list)

        # Simple replacement of local external agents.
        # DEF - need better recognition of external agent
        full_path = full_path.replace("./", "/")

        return full_path

    def reset(self):
        """
        Resets the origination tracking
        """
        self.tool_to_index_map = {}




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict

from neuro_san.internals.filters.message_filter import MessageFilter
from neuro_san.internals.messages.chat_message_type import ChatMessageType


class MaximalMessageFilter(MessageFilter):
    """
    MessageFilter implementation that lets everything through.
    """

    def allow_message(self, chat_message_dict: Dict[str, Any], message_type: ChatMessageType) -> bool:
        """
        Determine whether to allow the message through.

        :param chat_message_dict: The ChatMessage dictionary to process.
        :param message_type: The ChatMessageType of the chat_message_dictionary to process.
        :return: True if the message should be allowed through to the client. False otherwise.
        """
        # As long as the dictionary has some keys in it, we will pass it on.
        if any(chat_message_dict):
            return True

        return False



# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

import logging
from threading import Lock
from typing import Dict
from typing import Tuple
from watchdog.events import FileSystemEventHandler


class RegistryChangeHandler(FileSystemEventHandler):
    """
    Class for handling watchdog events in server registry directory.
    """
    MODIFIED = "modified"
    CREATED = "created"
    DELETED = "deleted"

    def __init__(self):
        """
        Constructor.
        """
        self.logger = logging.getLogger(self.__class__.__name__)
        self.lock: Lock = Lock()
        self.event_counters: Dict[str, int] =\
            {RegistryChangeHandler.MODIFIED: 0,
             RegistryChangeHandler.CREATED: 0,
             RegistryChangeHandler.DELETED: 0}

    def on_modified(self, event):
        """
        Handler for modified registry files.
        """
        self.handle_event(RegistryChangeHandler.MODIFIED, event.src_path)

    def on_created(self, event):
        """
        Handler for created registry files.
        """
        self.handle_event(RegistryChangeHandler.CREATED, event.src_path)

    def on_deleted(self, event):
        """
        Handler for deleted registry files.
        """
        self.handle_event(RegistryChangeHandler.DELETED, event.src_path)

    def handle_event(self, event_name: str, src_path: str):
        """
        Handle general watchdog event.
        """
        if not self.filter_src_name(src_path):
            return
        with self.lock:
            self.event_counters[event_name] += 1
        self.logger.info("🔔 File %s: %s", event_name, src_path)

    def reset_event_counters(self) -> Tuple[int, int, int]:
        """
        Reset event counters and return current counters.
        """
        with self.lock:
            modified: int = self.event_counters[self.MODIFIED]
            added: int = self.event_counters[self.CREATED]
            deleted: int = self.event_counters[self.DELETED]
            self.event_counters =\
                {self.MODIFIED: 0,
                 self.CREATED: 0,
                 self.DELETED: 0}
        return modified, added, deleted

    def filter_src_name(self, src_name: str) -> bool:
        """
        Filter source names we are getting notifications for.
        :param src_name: file path we get notified about
        :return: True if we should consider this file path;
                 False otherwise.
        """
        if not src_name:
            return False
        if src_name.endswith(".hocon") or src_name.endswith(".json"):
            return True
        return False




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict

from neuro_san.interfaces.coded_tool import CodedTool


class Calculator(CodedTool):
    """
    CodedTool implemetation of a calculator for the math_guy test.

    Upon activation by the agent hierarchy, a CodedTool will have its
    invoke() call called by the system.

    Implementations are expected to clean up after themselves.
    """

    async def async_invoke(self, args: Dict[str, Any], sly_data: Dict[str, Any]) -> Any:
        """
        Called when the coded tool is invoked asynchronously by the agent hierarchy.
        Strongly consider overriding this method instead of the "easier" synchronous
        version above when the possibility of making any kind of call that could block
        (like sleep() or a socket read/write out to a web service) is within the
        scope of your CodedTool.

        :param args: An argument dictionary whose keys are the parameters
                to the coded tool and whose values are the values passed for them
                by the calling agent.  This dictionary is to be treated as read-only.
        :param sly_data: A dictionary whose keys are defined by the agent hierarchy,
                but whose values are meant to be kept out of the chat stream.

                This dictionary is largely to be treated as read-only.
                It is possible to add key/value pairs to this dict that do not
                yet exist as a bulletin board, as long as the responsibility
                for which coded_tool publishes new entries is well understood
                by the agent chain implementation and the coded_tool implementation
                adding the data is not invoke()-ed more than once.
        :return: A return value that goes into the chat stream.
        """
        retval: float = 0.0

        operator: str = args.get("operator")
        if operator is None or not isinstance(operator, str):
            return "Don't understand non-string operators"

        x: float = sly_data.get("x")
        y: float = sly_data.get("y")

        if x is None or y is None:
            return "Need to set keys x and y in the sly_data as float operands"

        x = float(x)
        y = float(y)

        operator = operator.lower()
        if operator in ("add", "addition", "+", "plus"):
            retval = x + y
        elif operator in ("subtract", "subtraction", "-", "minus"):
            retval = x - y
        elif operator in ("multiply", "multiplication", "*", "times"):
            retval = x * y
        elif operator in ("divide", "division", "/", "over", "divided by"):
            if y != 0:
                retval = x / y
            else:
                return "Can't divide by 0"

        sly_data["equals"] = retval

        return "Check sly_data['equals'] for the result"




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import List
from typing import Type

import os

from langchain_core.callbacks.base import BaseCallbackHandler
from langchain_core.language_models.base import BaseLanguageModel

from leaf_common.config.dictionary_overlay import DictionaryOverlay
from leaf_common.config.resolver import Resolver
from leaf_common.parsers.dictionary_extractor import DictionaryExtractor

from neuro_san.internals.interfaces.context_type_llm_factory import ContextTypeLlmFactory
from neuro_san.internals.run_context.langchain.langchain_llm_factory import LangChainLlmFactory
from neuro_san.internals.run_context.langchain.llm_info_restorer import LlmInfoRestorer
from neuro_san.internals.run_context.langchain.standard_langchain_llm_factory import StandardLangChainLlmFactory


class DefaultLlmFactory(ContextTypeLlmFactory, LangChainLlmFactory):
    """
    Factory class for LLM operations

    Most methods take a config dictionary which consists of the following keys:

        "model_name"                The name of the model.
                                    Default if not specified is "gpt-3.5-turbo"

        "temperature"               A float "temperature" value with which to
                                    initialize the chat model.  In general,
                                    higher temperatures yield more random results.
                                    Default if not specified is 0.7

        "prompt_token_fraction"     The fraction of total tokens (not necessarily words
                                    or letters) to use for a prompt. Each model_name
                                    has a documented number of max_tokens it can handle
                                    which is a total count of message + response tokens
                                    which goes into the calculation involved in
                                    get_max_prompt_tokens().
                                    By default the value is 0.5.

        "max_tokens"                The maximum number of tokens to use in
                                    get_max_prompt_tokens(). By default this comes from
                                    the model description in this class.
    """

    def __init__(self):
        """
        Constructor
        """
        self.llm_infos: Dict[str, Any] = {}
        self.overlayer = DictionaryOverlay()
        self.llm_factories: List[LangChainLlmFactory] = [
            StandardLangChainLlmFactory()
        ]

    def load(self, agent_llm_info_file: str):
        """
        Loads the LLM information from hocon files.
        """
        restorer = LlmInfoRestorer()
        self.llm_infos = restorer.restore()

        # Mix in user-specified llm info, if available.
        # First check "agent_llm_info_file" key from agent network hocon.
        # If that is unavailable, fallback to env variable.
        llm_info_file: str = agent_llm_info_file
        if not agent_llm_info_file:
            llm_info_file = os.getenv("AGENT_LLM_INFO_FILE")
        if llm_info_file is not None and len(llm_info_file) > 0:
            extra_llm_infos: Dict[str, Any] = restorer.restore(file_reference=llm_info_file)
            self.llm_infos = self.overlayer.overlay(self.llm_infos, extra_llm_infos)

        # sanitize the llm_infos keys
        self.llm_infos = self.sanitize_keys(self.llm_infos)

        # Resolve any new llm factories
        extractor = DictionaryExtractor(self.llm_infos)
        llm_factory_classes: List[str] = []
        llm_factory_classes = extractor.get("classes.factories", llm_factory_classes)
        if not isinstance(llm_factory_classes, List):
            raise ValueError(f"The classes.factories key in {llm_info_file} must be a list of strings")

        for llm_factory_class_name in llm_factory_classes:
            llm_factory: LangChainLlmFactory = self.resolve_one_llm_factory(llm_factory_class_name, llm_info_file)
            # Success. Tack it on to the list
            self.llm_factories.append(llm_factory)

    def resolve_one_llm_factory(self, llm_factory_class_name: str, llm_info_file: str) -> LangChainLlmFactory:
        """
        :param llm_factory_class_name: A single class name to resolve.
        :param llm_info_file: The name of the hocon file with the class names, to reference
                        when exceptions are thrown.
        :return: A LangChainLlmFactory instance as per the input
        """
        if not isinstance(llm_factory_class_name, str):
            raise ValueError(f"The value for the classes.factories key in {llm_info_file} "
                             "must be a list of strings")

        class_split: List[str] = llm_factory_class_name.split(".")
        if len(class_split) <= 2:
            raise ValueError(f"Value in the classes.factories in {llm_info_file} must be of the form "
                             "<package_name>.<module_name>.<ClassName>")

        # Create a list of a single package given the name in the value
        packages: List[str] = [".".join(class_split[:-2])]
        class_name: str = class_split[-1]
        resolver = Resolver(packages)

        # Resolve the class name
        llm_factory_class: Type[LangChainLlmFactory] = None
        try:
            llm_factory_class: Type[LangChainLlmFactory] = \
                resolver.resolve_class_in_module(class_name, module_name=class_split[-2])
        except AttributeError as exception:
            raise ValueError(f"Class {llm_factory_class_name} in {llm_info_file} "
                             "not found in PYTHONPATH") from exception

        # Instantiate it
        try:
            llm_factory: LangChainLlmFactory = llm_factory_class()
        except TypeError as exception:
            raise ValueError(f"Class {llm_factory_class_name} in {llm_info_file} "
                             "must have a no-args constructor") from exception

        # Make sure its the correct type
        if not isinstance(llm_factory, LangChainLlmFactory):
            raise ValueError(f"Class {llm_factory_class_name} in {llm_info_file} "
                             "must be of type LangChainLlmFactory")
        return llm_factory

    def create_llm(self, config: Dict[str, Any], callbacks: List[BaseCallbackHandler] = None) -> BaseLanguageModel:
        """
        Creates a langchain LLM based on the 'model_name' value of
        the config passed in.

        :param config: A dictionary which describes which LLM to use.
                See the class comment for details.
        :param callbacks: A list of BaseCallbackHandlers to add to the chat model.
        :return: A BaseLanguageModel (can be Chat or LLM)
                Can raise a ValueError if the config's model_name value is
                unknown to this method.
        """
        full_config: Dict[str, Any] = self.create_full_llm_config(config)
        llm: BaseLanguageModel = self.create_base_chat_model(full_config, callbacks)
        return llm

    def create_full_llm_config(self, config: Dict[str, Any]) -> Dict[str, Any]:
        """
        :param config: The llm_config from the user
        :return: The fully specified config with defaults filled in.
        """
        default_config: Dict[str, Any] = self.llm_infos.get("default_config")
        use_config = self.overlayer.overlay(default_config, config)

        model_name = use_config.get("model_name")

        llm_entry = self.llm_infos.get(model_name)
        if llm_entry is None:
            raise ValueError(f"No llm entry for model_name {model_name}")

        # Get some bits from the llm_entry
        use_model_name = llm_entry.get("use_model_name", model_name)
        if len(llm_entry.keys()) <= 2 and use_model_name is not None:
            # We effectively have an alias. Switch out the llm entry.
            llm_entry = self.llm_infos.get(use_model_name)
            if llm_entry is None:
                raise ValueError(f"No llm entry for use_model_name {use_model_name} in {model_name}")

        # Take a look at the chat classes.
        chat_class_name: str = llm_entry.get("class")
        if chat_class_name is None:
            raise ValueError(f"llm info entry for {use_model_name} requires a 'class' key/value pair.")

        # Get defaults for the chat class
        chat_args: Dict[str, Any] = self.get_chat_class_args(chat_class_name, use_model_name)

        # Get a new sense of the default config now that we have the default args for the chat class.
        default_config = self.overlayer.overlay(chat_args, default_config)

        # Now that we have the true defaults, overlay the config that came in to get the
        # config we are going to use.
        full_config: Dict[str, Any] = self.overlayer.overlay(default_config, config)
        full_config["class"] = chat_class_name
        full_config["model_name"] = llm_entry.get("use_model_name", use_model_name)

        # Attempt to get a max_tokens through calculation
        full_config["max_tokens"] = self.get_max_prompt_tokens(full_config)

        return full_config

    def get_chat_class_args(self, chat_class_name: str, use_model_name: str) -> Dict[str, Any]:
        """
        :param chat_class_name: string name of the chat class to look up.
        :param use_model_name: the original model name that prompted the chat class lookups
        :return: A dictionary of default arguments for the chat class.
                Can throw an exception if the chat class does not exist.
        """

        # Find the chat class.
        chat_classes: Dict[str, Any] = self.llm_infos.get("classes")
        chat_class: Dict[str, Any] = chat_classes.get(chat_class_name)
        if chat_class is None:
            raise ValueError(f"llm info entry for {use_model_name} uses a 'class' of {chat_class_name} "
                             "which is not defined in the 'classes' table.")

        # Get the args from the chat class
        args: Dict[str, Any] = chat_class.get("args")

        extends: str = chat_class.get("extends")
        if extends is not None:
            # If this class extends another, get its args too.
            extends_args: Dict[str, Any] = self.get_chat_class_args(extends, use_model_name)
            args = self.overlayer.overlay(args, extends_args)

        return args

    def create_base_chat_model(self, config: Dict[str, Any],
                               callbacks: List[BaseCallbackHandler] = None) -> BaseLanguageModel:
        """
        Create a BaseLanguageModel from the fully-specified llm config.
        :param config: The fully specified llm config which is a product of
                    _create_full_llm_config() above.
        :param callbacks: A list of BaseCallbackHandlers to add to the chat model.
        :return: A BaseLanguageModel (can be Chat or LLM)
                Can raise a ValueError if the config's class or model_name value is
                unknown to this method.
        """
        llm: BaseLanguageModel = None

        # Loop through the loaded factories in order until we can find one
        # that can create the llm.
        found_exception: Exception = None
        for llm_factory in self.llm_factories:
            try:
                llm = llm_factory.create_base_chat_model(config, callbacks)
                if llm is not None and isinstance(llm, BaseLanguageModel):
                    # We found what we were looking for
                    found_exception = None
                    break
            except ValueError as exception:
                # Let the next model have a crack
                found_exception = exception

        if found_exception is not None:
            raise found_exception

        return llm

    def get_max_prompt_tokens(self, config: Dict[str, Any]) -> int:
        """
        :param config: A dictionary which describes which LLM to use.
        :return: The maximum number of tokens given the 'model_name' in the
                config dictionary.
        """

        model_name = config.get("model_name")

        llm_entry = self.llm_infos.get(model_name)
        if llm_entry is None:
            raise ValueError(f"No llm entry for model_name {model_name}")

        use_model_name = llm_entry.get("use_model_name", model_name)
        if len(llm_entry.keys()) <= 2 and use_model_name is not None:
            # We effectively have an alias. Switch out the llm entry.
            llm_entry = self.llm_infos.get(use_model_name)

        entry_max_tokens = llm_entry.get("max_output_tokens")
        prompt_token_fraction = config.get("prompt_token_fraction")
        use_max_tokens = int(entry_max_tokens * prompt_token_fraction)

        # Allow the actual value for max_tokens to come from the config, if there
        max_prompt_tokens = config.get("max_tokens", use_max_tokens)
        if max_prompt_tokens is None:
            max_prompt_tokens = use_max_tokens

        return max_prompt_tokens

    def strip_outer_quotes(self, s: str) -> str:
        """
        :param s: The input string to sanitize.
        :return: The input string without surrounding multiple quotes, if they were present.
        Otherwise, returns the string unchanged.
        """
        if len(s) >= 2 and s[0] in ["'", '"']:
            quote_char = s[0]
            unquoted = s[1:-1]

            # Only strip quotes if inner value does not contain the same quote char
            # and does not start with whitespace
            if quote_char not in unquoted and not unquoted.startswith(" "):
                return unquoted
        return s

    def sanitize_keys(self, d: Dict[str, Any]) -> Dict[str, Any]:
        """
        Returns a new dictionary with surrounding double quotes stripped from all keys.
        This is typically used to clean up configuration dictionaries where key names may
        be quoted (e.g., '"llama3.1"' instead of "llama3.1") due to parsing artifacts.
        Only the top-level keys are sanitized; nested keys are left unchanged.
        :param d: The input dictionary with potentially quoted keys.
        :return: A new dictionary with the same values, but with sanitized keys.
        """
        return {self.strip_outer_quotes(k): v for k, v in d.items()}
